from __future__ import generator_stop

import errno
import logging
import os
import os.path
import platform
import re
import shutil
import stat
from fnmatch import fnmatch
from functools import partial
from itertools import zip_longest
from operator import attrgetter
from os import DirEntry, PathLike, fspath
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Iterable, Iterator, List, Optional, Set, Tuple, Union

from ._files import BaseDirEntry, MyDirEntryT, entrysuffix
from .datetime import datetime_from_utc_timestamp
from .file import FILE_IO_BUFFER_SIZE, equal_files, iterfilelike
from .iter import is_empty
from .ops import logical_implication
from .os import _not_available, islink, uncabspath

if TYPE_CHECKING:
    from datetime import datetime

PathType = Union[str, PathLike]
EntryType = Union[Path, DirEntry]

logger = logging.getLogger(__name__)

ascii_control = {chr(i) for i in range(1, 32)}
ntfs_illegal_chars = {"\0", "/"}
fat_illegal_chars = {"\0", "/", "?", "<", ">", "\\", ":", "*", "|", '"', "^"}  # ^ really?

windows_ntfs_illegal_chars = ntfs_illegal_chars | {"\\", ":", "*", "?", '"', "<", ">", "|"} | ascii_control
windows_fat_illegal_chars = fat_illegal_chars

windows_illegal_chars = windows_ntfs_illegal_chars | windows_fat_illegal_chars
unix_illegal_chars = {"\0", "/"}
mac_illegal_chars = {"\0", "/", ":"}

windows_sep = r"\/"
linux_sep = r"/"
mac_sep = r"/:"

windows_reserved_names = {
    "CON",
    "PRN",
    "AUX",
    "NUL",
    "COM1",
    "COM2",
    "COM3",
    "COM4",
    "COM5",
    "COM6",
    "COM7",
    "COM8",
    "COM9",
    "LPT1",
    "LPT2",
    "LPT3",
    "LPT4",
    "LPT5",
    "LPT6",
    "LPT7",
    "LPT8",
    "LPT9",
}


class fileextensions:
    audio = ("wav", "mp3", "aac", "flac", "m4a", "m4b", "aiff", "ogg", "wma", "mka", "ac3", "dts")
    video = (
        "avi",
        "mp4",
        "m4v",
        "mkv",
        "mpg",
        "mpeg",
        "wmv",
        "mov",
        "flv",
        "f4v",
        "webm",
        "vob",
        "ogv",
        "m2ts",
        "rm",
        "rmvb",
        "asf",
        "3gp",
        "nsv",
        "divx",
        "ogm",
    )
    images = ("bmp", "jpg", "jpeg", "png", "gif", "tif", "tiff", "tga")

    archives = ("zip", "rar", "ace", "tar", "tgz", "tbz", "7z", "cab", "dmg", "wim")
    bytecode = ("pyc", "class")
    configuration = ("ini", "toml", "conf", "reg")
    database = ("cat", "sqlite", "db")
    documents = ("txt", "pdf", "doc", "docx", "ps", "rtf", "pgs")
    executables = ("exe", "dll", "so", "scr", "msi", "pyd", "sys", "cpl")  # native code
    image_archives = ("cbz", "cbr", "cb7", "cbt", "cba")
    scripts = ("bat", "ps1", "sh", "c", "cpp", "cs", "py", "php", "js", "lua", "pl")  # and source code files

    disc_images_binary = ("iso", "bin", "mdf", "img")
    disc_images_sidecar = ("cue", "mds", "ccd")
    disc_images = disc_images_binary + disc_images_sidecar

    game_images = ("gcn", "nds", "3ds", "wad", "xbx", "gba", "gb", "nes", "sfc", "n64", "z64")
    compressed = ("gz", "bz2", "lzma", "xz", "z")
    subtitles = ("srt", "sub", "idx", "sup", "ssa", "ass")
    scene = ("nfo", "sfv")

    windows = ("library-ms", "desklink", "lnk", "dmp")
    macos = ("crash",)
    os = windows + macos


class IllegalFilename(ValueError):
    pass


class WindowsIllegalFilename(IllegalFilename):
    pass


class FileProperties:

    __slots__ = ("relpath", "size", "isdir", "abspath", "id", "modtime", "hash")

    def __init__(self, relpath, size, isdir, abspath=None, id=None, modtime=None, hash=None):
        # type: (str, int, bool, Optional[str], Optional[Any], Optional[datetime], Optional[str]) -> None

        self.relpath = relpath
        self.size = size
        self.isdir = isdir
        self.abspath = abspath
        self.id = id
        self.modtime = modtime
        self.hash = hash

    @classmethod
    def keys(cls):
        return cls.__slots__

    def values(self):
        return attrgetter(*self.__slots__)(self)

    def __iter__(self):
        # type: () -> tuple

        return iter(self.values())

    def __eq__(self, other):
        # type: (Any, ) -> bool

        return self.values() == other.values()

    def __repr__(self):
        # type: () -> str

        args = (f"{k}={v!r}" for k, v in zip(self.keys(), self.values()) if v is not None)
        return "FileProperties({})".format(", ".join(args))


class DirEntryStub:
    __slots__ = ("name", "path")

    def __init__(self, name, path):
        self.name = name
        self.path = path


class MyDirEntry(BaseDirEntry):
    __slots__ = ("basepath", "_relpath", "follow")

    def __init__(self, entry):
        BaseDirEntry.__init__(self, entry)

        self.basepath = None  # type: Optional[str]
        self._relpath = None  # type: Optional[str]
        self.follow = True

    @property
    def relpath(self):
        # type: () -> str

        if self._relpath is not None:
            return self._relpath

        if self.basepath is None:
            raise RuntimeError("relpath cannot be returned, because basepath is not set")

        self._relpath = os.path.relpath(self.entry.path, self.basepath)
        return self._relpath

    @relpath.setter
    def relpath(self, path):
        # type: (str, ) -> None

        self._relpath = path


def mdatetime(path, aslocal=False):
    # type: (PathType, bool) -> datetime

    """Returns the last modified date of `path`
    as a timezone aware datetime object.

    If `aslocal=True` it will be formatted as local time,
    and UTC otherwise (the default).
    """

    if isinstance(path, (Path, DirEntry)):
        mtime = path.stat().st_mtime
    else:
        mtime = os.stat(path).st_mtime

    return datetime_from_utc_timestamp(mtime, aslocal)


def rename(old, new):
    # type: (PathType, PathType) -> None

    """Renames `old` to `new`. Fails if `new` already exists.
    This is the default behaviour of `os.rename` on Windows.
    This function should do the same cross-platform.
    It is however not race free.
    """

    # fixme: renaming a non-existing file in a non-existing folder yields a PermissionError not FileNotFoundError

    if os.path.exists(new):
        raise FileExistsError(new)

    os.renames(old, new)  # fixme: don't use rename*s*


def copy_file_generator(source, dest, buffer_size=FILE_IO_BUFFER_SIZE, overwrite_readonly=False):
    # type: (PathType, PathType, int, bool) -> Iterator[None]

    """Partial file is not deleted if exception gets raised anywhere."""

    try:
        os.makedirs(os.path.dirname(dest))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    if overwrite_readonly:
        try:
            make_writeable(dest)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    with open(source, "rb") as src, open(dest, "wb") as dst:
        for data in iterfilelike(src, chunk_size=buffer_size):
            dst.write(data)
            yield

    shutil.copystat(source, dest)


def st_mode_to_str(st_mode):
    # type: (int, ) -> str

    if stat.S_ISDIR(st_mode):
        return "directory"
    elif stat.S_ISREG(st_mode):
        return "regular file"
    elif stat.S_ISCHR(st_mode):
        return "character special device file"
    elif stat.S_ISBLK(st_mode):
        return "block special device file"
    elif stat.S_ISFIFO(st_mode):
        return "named pipe"
    elif stat.S_ISLNK(st_mode):
        return "symbolic link"
    elif stat.S_ISSOCK(st_mode):
        return "socket"
    else:
        return "unknown"


def append_to_filename(path, s):
    # type: (PathType, str) -> str

    """Add `s` to the filename given in `path`. It's added before the extension, not after."""

    root, ext = os.path.splitext(path)
    return root + s + ext


def scandir_error_log(entry, exception):
    logger.exception("Error in %s", entry.path, exc_info=exception)


def scandir_error_raise(entry, exception):
    raise exception


def scandir_error_ignore(entry, exception):
    pass


def _scandir_rec_skippable(rootentry, files=True, others=False, follow_symlinks=True, errorfunc=scandir_error_raise):
    # type: (DirEntry, bool, bool, bool, Callable[[DirEntry, Exception], None]) -> Iterator[MyDirEntry]

    try:
        with os.scandir(rootentry.path) as it:
            for entry in it:
                if files and entry.is_file(follow_symlinks=follow_symlinks):
                    yield MyDirEntry(entry)
                elif entry.is_dir(follow_symlinks=follow_symlinks):
                    if not follow_symlinks and islink(entry):  # must be a windows junction
                        continue
                    se = MyDirEntry(entry)
                    yield se
                    if se.follow:
                        for e in _scandir_rec_skippable(entry, files, others, follow_symlinks, errorfunc):
                            yield e
                else:
                    if others:
                        yield MyDirEntry(entry)
    except OSError as e:
        errorfunc(rootentry, e)


def _scandir_rec(
    rootentry, files=True, dirs=False, others=False, rec=True, follow_symlinks=True, errorfunc=scandir_error_raise
):
    # type: (DirEntry, bool, bool, bool, bool, bool, Callable[[DirEntry, Exception], None]) -> Iterator[DirEntry]

    try:
        with os.scandir(rootentry.path) as it:
            for entry in it:
                if files and entry.is_file(follow_symlinks=follow_symlinks):
                    yield entry
                elif entry.is_dir(follow_symlinks=follow_symlinks):
                    if not follow_symlinks and islink(
                        entry
                    ):  # must be a windows junction. entry.is_symlink() doesn't support junctions
                        continue
                    if dirs:
                        yield entry
                    if rec:
                        for e in _scandir_rec(entry, files, dirs, others, rec, follow_symlinks, errorfunc):
                            yield e
                else:
                    if others:
                        yield entry
    except OSError as e:
        errorfunc(rootentry, e)


def scandir_rec(
    path,
    files=True,
    dirs=False,
    others=False,
    rec=True,
    follow_symlinks=True,
    relative=False,
    allow_skip=False,
    errorfunc=scandir_error_log,
):
    # type: (PathType, bool, bool, bool, bool, bool, bool, bool, Callable[[MyDirEntryT, Exception], None]) -> Iterator[MyDirEntryT]

    if not logical_implication(allow_skip, dirs and rec):
        raise ValueError("allow_skip implies dirs and rec")

    if isinstance(path, PathLike):
        path = fspath(path)

    entry = DirEntryStub(os.path.basename(path), uncabspath(path))  # for python 2 compat. and long filename support

    if not allow_skip:
        it = _scandir_rec(entry, files, dirs, others, rec, follow_symlinks, errorfunc)
    else:
        it = _scandir_rec_skippable(entry, files, others, follow_symlinks, errorfunc)

    if not relative:
        return it
    else:
        basepath = uncabspath(path)
        if not allow_skip:
            # entry is a `os.DirEntry`
            def modpathrelative(entry):
                entry = MyDirEntry(entry)
                entry.basepath = basepath
                return entry

        else:
            # entry is a `MyDirEntry`
            def modpathrelative(entry):
                entry.basepath = basepath
                return entry

        return map(modpathrelative, it)


def scandir_ext(
    path: PathType,
    extensions: Set[str],
    rec: bool = True,
    follow_symlinks: bool = False,
    relative: bool = False,
    errorfunc: Callable = scandir_error_log,
) -> Iterator[DirEntry]:

    for entry in scandir_rec(
        path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks, relative=relative, errorfunc=errorfunc
    ):
        if entrysuffix(entry).lower() in extensions:
            yield entry


# fixme: benchmark and delete
def scandir_ext_2(path: str, extensions: Set[str], rec: bool = True) -> Iterator[Path]:

    if rec:
        paths = Path(uncabspath(path)).rglob("*")
    else:
        paths = Path(uncabspath(path)).glob("*")

    for p in paths:
        if p.suffix.lower() in extensions:
            yield p


def _scandir_depth(rootentry, depth, errorfunc):

    try:
        with os.scandir(rootentry.path) as it:
            for entry in it:
                if entry.is_dir():
                    yield depth, entry
                    yield from _scandir_depth(entry, depth + 1, errorfunc)

        with os.scandir(rootentry.path) as it:
            for entry in it:
                if entry.is_file():
                    yield depth, entry

    except OSError as e:
        errorfunc(rootentry, e)


def scandir_depth(path, depth=0, onerror=scandir_error_log):
    # type: (str, int, Callable) -> Iterator[Tuple[int, DirEntry]]

    """Recursive version of `scandir` which yields the current path depth
    along with the DirEntry.
    Directories are returned first, then files.
    The order is arbitrary otherwise.
    """

    entry = DirEntryStub(os.path.basename(path), uncabspath(path))  # for python 2 compat. and long filename support
    return _scandir_depth(entry, depth, onerror)


def make_readonly(path, stats=None):
    # type: (PathType, Optional[os.stat_result]) -> None

    """deletes all write flags"""
    if not stats:
        stats = os.stat(path)
    os.chmod(path, stat.S_IMODE(stats.st_mode) & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)


def make_writeable(path, stats=None):
    # type: (PathType, Optional[os.stat_result]) -> None

    """set owner write flag"""
    if not stats:
        stats = os.stat(path)
    os.chmod(path, stat.S_IMODE(stats.st_mode) | stat.S_IWRITE)


def is_writeable(stats):
    # type: (os.stat_result, ) -> bool

    return stats.st_mode & stat.S_IWRITE != 0


def is_readable(stats):
    # type: (os.stat_result, ) -> bool

    return stats.st_mode & stat.S_IREAD != 0


def isfile(stats):
    # type: (os.stat_result, ) -> bool

    return stat.S_ISREG(stats.st_mode)


def isdir(stats):
    # type: (os.stat_result, ) -> bool

    return stat.S_ISDIR(stats.st_mode)


# was: islink
def statsislink(stats):
    # type: (os.stat_result, ) -> bool

    return stat.S_ISLNK(stats.st_mode)


def extract_basic_stat_info(stats):
    # type: (os.stat_result, ) -> tuple

    """Returns size, modification, creation time and access mode."""
    return (stats.st_size, stats.st_mtime, stats.st_ctime, stats.st_mode)


def safe_filename(filename, replacement=""):
    # type: (str, str) -> str

    WIN = windows_illegal_chars
    UNIX = unix_illegal_chars
    MAC = mac_illegal_chars
    bad = set.union(WIN, UNIX, MAC)

    safe_filename_translation_table = str.maketrans({c: replacement for c in bad})

    return filename.translate(safe_filename_translation_table)  # fixme: return callable which accepts only filename


safe_filename_simple = safe_filename


def _char_subber(s, illegal_chars, replacement):
    # type: (str, Set[str], str) -> str

    if set(replacement) & illegal_chars:
        raise ValueError("replace character cannot be a illegal character")

    regex = "[" + re.escape("".join(illegal_chars)) + "]"
    return re.sub(regex, replacement, s)


def _char_splitter(s, chars):
    # type: (str, Iterable[str]) -> List[str]

    regex = "[" + re.escape("".join(chars)) + "]"
    return re.split(regex, s)


def _special_subber(s, replacement="_"):
    # type: (str, str) -> str

    if s == "." or s == "..":
        return s.replace(".", replacement)

    return s


def windows_compliant_filename(filename, replacement="_"):
    # type: (str, str) -> str
    """https://msdn.microsoft.com/en-us/library/aa365247.aspx"""
    # reg. wikipedia: \x7f is valid!
    # trailing dots and spaces are ignored by windows

    ret = _char_subber(filename, windows_illegal_chars, replacement).rstrip(". ")  # strip or replace?, translate

    root, _ = os.path.splitext(ret)

    if root.upper() in windows_reserved_names:
        raise WindowsIllegalFilename("filename would result in reserved name")

    return ret


def windows_compliant_dirname(filename, replacement):
    # type: (str, str) -> str
    """https://msdn.microsoft.com/en-us/library/aa365247.aspx"""
    # reg. wikipedia: \x7f is valid!
    # trailing dots and spaces are ignored by windows

    ret = _char_subber(filename, windows_illegal_chars, replacement).rstrip(" ")

    root, _ = os.path.splitext(ret)

    if root.upper() in windows_reserved_names:
        raise WindowsIllegalFilename("filename would result in reserved name")

    return ret


def linux_compliant_filename(filename, replacement="_"):
    # type: (str, str) -> str

    filename = _special_subber(filename, replacement)
    return _char_subber(filename, unix_illegal_chars, replacement)


def linux_compliant_dirname(filename, replacement="_"):
    # type: (str, str) -> str

    return _char_subber(filename, unix_illegal_chars, replacement)


def mac_compliant_filename(filename, replacement="_"):
    # type: (str, str) -> str

    filename = _special_subber(filename, replacement)
    return _char_subber(filename, mac_illegal_chars, replacement)


def mac_compliant_dirname(filename, replacement="_"):
    # type: (str, str) -> str

    return _char_subber(filename, mac_illegal_chars, replacement)


def windows_split_path(s):
    # type: (str, ) -> List[str]

    return _char_splitter(s, windows_sep)


def linux_split_path(s):
    # type: (str, ) -> List[str]

    return _char_splitter(s, linux_sep)


def mac_split_path(s):
    # type: (str, ) -> List[str]

    return _char_splitter(s, mac_sep)


system = platform.system()

if system == "Windows":
    compliant_filename = windows_compliant_filename
    compliant_dirname = windows_compliant_dirname
    split_path = windows_split_path
elif system == "Linux":
    compliant_filename = linux_compliant_filename
    compliant_dirname = linux_compliant_dirname
    split_path = linux_split_path
elif system == "Darwin":
    compliant_filename = mac_compliant_filename
    compliant_dirname = mac_compliant_dirname
    split_path = mac_split_path
else:
    compliant_filename = _not_available("compliant_filename")
    compliant_dirname = _not_available("compliant_dirname")
    split_path = _not_available("split_path")


def compliant_path(path, replace="_"):
    # type: (str, str) -> str

    fn_func = partial(compliant_dirname, replace=replace)
    path_compontents = split_path(path)

    ret = os.sep.join(map(fn_func, path_compontents[:-1]))
    ret = os.path.join(ret, compliant_filename(path_compontents[-1]))
    return ret


def reldirname(path):
    # type: (PathType, ) -> str

    ret = os.path.dirname(path)
    if not ret:
        ret = "."

    return ret


def equal_dirs_iter(dir1, dir2, follow_symlinks=False):

    """Tests if two directories are equal. Doesn't handle links."""

    def entry_path_size(entry):
        return entry.path, entry.stat().st_size

    files1 = sorted(map(entry_path_size, scandir_rec(dir1, follow_symlinks=follow_symlinks)))  # get rel paths
    files2 = sorted(map(entry_path_size, scandir_rec(dir2, follow_symlinks=follow_symlinks)))  # get rel paths

    for (path1, size1), (path2, size2) in zip_longest(files1, files2):
        if path1 != path2:
            yield ("name", path1, path2)

        elif size1 != size2:
            yield ("size", path1, path2)

    for (path1, size1), (path2, size2) in zip_longest(files1, files2):
        if path1 == path2 and size1 == size2 and not equal_files(path1, path2):
            yield ("data", path1, path2)


def equal_dirs(dir1, dir2):
    return is_empty(equal_dirs_iter(dir1, dir2))


def realpath_win(path):
    # type: (PathType, ) -> str

    """fix for os.path.realpath() which doesn't work under Win7"""

    path = os.path.abspath(path)
    if islink(path):
        return os.path.normpath(os.path.join(os.path.dirname(path), os.readlink(path)))
    else:
        return path


def search(directories, pattern, dirs=True, files=True, rec=True, follow_symlinks=False):
    # type: (Iterable[PathType], str, bool, bool, bool, bool) -> Iterator[DirEntry]

    """Search for files and folders matching the wildcard `pattern`."""

    for directory in directories:
        for entry in scandir_rec(directory, dirs=dirs, files=files, rec=rec, follow_symlinks=follow_symlinks):
            if fnmatch(entry.name, pattern):
                yield entry


def rename_files_in_folder(path, pattern, transform, template, rec=True, follow_symlinks=False):
    # type: (PathType, str, Callable, str, bool, bool) -> None

    cpattern = re.compile(pattern)

    for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
        filepath = entry.path
        base, name = os.path.split(filepath)

        m = cpattern.match(name)
        if not m:
            continue
        args = transform(*m.groups())

        to = template.format(*args)
        os.rename(filepath, os.path.join(base, to))


def clean_directory_by_extension(path, ext, rec=True, follow_symlinks=False):
    # type: (PathType, str, bool, bool) -> None

    """Deletes all files of type `ext` if the same file without the ext exists
    and rename if it does not exit."""

    # delete .ext if normal file exists
    for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
        filepath = entry.path
        if not filepath.endswith(ext):
            file_to_delete = filepath + ext
            if os.path.isfile(file_to_delete):
                logger.info("Remove: %s", file_to_delete)
                try:
                    os.remove(file_to_delete)
                except OSError:
                    logger.exception("Could not delete: %s", file_to_delete)
                    break

    # rename all .ext to normal
    for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
        filepath = entry.path
        if entry.name.endswith(ext):
            target = filepath[: -len(ext)]
            try:
                logger.info("Rename: %s -> %s", filepath, target)
                os.rename(filepath, target)
            except OSError:
                logger.exception("Should not have happened")
                break


def iter_links(path):
    # type: (PathType, ) -> Iterator[str]

    """Yields all directory symlinks (and junctions on windows)."""

    for dirpath, dirnames, filenames in os.walk(path):
        for dirname in dirnames:
            joined = os.path.join(dirpath, dirname)
            if islink(joined):
                yield joined


def normalize_seps(path):
    # type: (str, ) -> str

    """Converts \\ to /"""

    return path.replace("\\", "/")


class Counts:
    __slots__ = ("dirs", "files", "others")

    def __init__(self):
        self.dirs = 0
        self.files = 0
        self.others = 0

    def __iadd__(self, other):
        # type: (Counts, ) -> Counts

        self.dirs += other.dirs
        self.files += other.files
        self.others += other.others

        return self

    def null(self):
        # type: () -> bool

        return self.dirs == 0 and self.files == 0 and self.others == 0


def _scandir_counts(rootentry, files=True, others=True, rec=True, total=False, errorfunc=scandir_error_raise):
    # type: (MyDirEntryT, bool, bool, bool, bool, Callable) -> Iterator[Tuple[DirEntry, Optional[Counts]]]

    counts = Counts()

    try:
        with os.scandir(rootentry.path) as it:
            for entry in it:

                if entry.is_dir():
                    counts.dirs += 1
                    if rec:
                        for subentry, subcounts in _scandir_counts(entry, files, others, rec, total, errorfunc):
                            yield subentry, subcounts

                            if total:
                                assert subcounts
                                counts += subcounts

                elif entry.is_file():
                    counts.files += 1
                    if files:
                        yield entry, None

                else:
                    counts.others += 1
                    if others:
                        yield entry, None

        yield rootentry, counts  # yield after the loop

    except OSError as e:
        errorfunc(rootentry, e)


def scandir_counts(path, files=True, others=True, rec=True, total=False, onerror=scandir_error_log):
    # type: (PathType, bool, bool, bool, bool, Callable) -> Iterator[Tuple[DirEntry, Optional[Counts]]]

    """A recursive variant of scandir() which also returns the number of files/directories
    within directories.
    If total is True, the numbers will be calculated recursively as well.
    """

    if isinstance(path, PathLike):
        path = fspath(path)

    entry = DirEntryStub(os.path.basename(path), uncabspath(path))
    return _scandir_counts(entry, files, others, rec, total, onerror)


def shutil_onerror_remove_readonly(func, path, exc_info):

    """onerror function for the `shutil.rmtree` function.
    Makes read-only files writable and tries again.
    """

    stats = os.stat(path)
    if is_readable(stats):
        make_writeable(path, stats)
        func(path)
    else:
        raise



from __future__ import generator_stop

from functools import wraps
from warnings import warn


def deprecated(msg, stacklevel=2):
    def decorator(func):
        @wraps(func)
        def inner(*args, **kwargs):
            warn(msg, DeprecationWarning, stacklevel)
            print("DeprecationWarning:", msg)
            return func(*args, **kwargs)

        return inner

    return decorator



from __future__ import generator_stop

import platform
from collections import namedtuple

_usagetuple = namedtuple("usage", "total used free")
_volumeinfotuple = namedtuple(
    "volumeinfo", "VolumeName VolumeSerialNumber MaximumComponentLength FileSystemFlags FileSystemName"
)


def is_os_64bit():
    return platform.machine().endswith("64")



from __future__ import generator_stop

from collections import Counter
from itertools import chain
from math import exp, log2
from sys import stderr
from typing import TYPE_CHECKING

import numpy as np

from .datastructures.sparse_matrix import VariableRowMatrix
from .encoder import BatchLabelEncoder
from .iter import progress
from .numpy import batchtopk, categorical
from .object import cast

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, Iterator, List, MutableMapping, Optional, Tuple, Union

    from .typing import SizedIterable

    RawDocument = List[int]
    IterableDocuments = Iterable[SizedIterable[int]]
    TopicsMapping = MutableMapping[Tuple[int, int], int]
    Indices = Union[slice, np.ndarray]


def top_topics(id2word, term_topic_matrix, num_words=10, decimals=2):
    # type: (Dict[int, str], np.ndarray, int, int) -> Iterator[List[Tuple[str, float]]]

    """
    term_topic_matrix: float[K, V]
    """

    indices_list, probs_list = batchtopk(term_topic_matrix, num_words, reverse=True)
    np.around(probs_list, decimals=decimals, out=probs_list)

    for indices, probs in zip(indices_list, probs_list):
        yield [(id2word[id], prob) for id, prob in zip(indices, probs)]


def format_topics(topics, linesep="\n", tokensep="\t"):
    # type: (Iterable[List[Tuple[str, float]]], str, str) -> str

    buffer = []

    for terms in topics:
        buffer.append(tokensep.join(f"({term},{prob})" for term, prob in terms))

    return linesep.join(buffer)


class LDADocument(SizedIterable[int]):
    def __init__(self, words):
        # type: (RawDocument, ) -> None

        self.words = words

    def __iter__(self):
        # type: () -> Iterator[int]

        return iter(self.words)

    def __len__(self):
        # type: () -> int

        return len(self.words)

    def __repr__(self):
        # type: () -> str

        return f"LDADocument({repr(self.words)})"


class LDABase:

    K: int
    nmk: np.ndarray
    nkt: np.ndarray
    nk: np.ndarray
    nm: np.ndarray

    α: float
    β: float
    αsum: float
    βsum: float

    docs: List[LDADocument]

    def __init__(self, seed=None):
        # type: (Optional[int], ) -> None

        np.seterr(all="raise")
        np.random.seed(seed)

    def init_topic(self):
        # type: () -> int

        return np.random.randint(self.K)

    def add_counts(self, m, t, k):
        # type: (int, int, int) -> None

        self.nmk[m, k] += 1
        self.nkt[k, t] += 1
        self.nk[k] += 1

    def sub_counts(self, m, t, k):
        # type: (int, int, int) -> None

        self.nmk[m, k] -= 1
        self.nkt[k, t] -= 1
        self.nk[k] -= 1

    def _validate(self, M, K, V):
        # type: (int, int, int) -> None

        if M <= 0:
            raise ValueError("Number of documents must be larger than zero")
        if K <= 0:
            raise ValueError("Number of topics must be larger than zero")
        if V <= 0:
            raise ValueError("Vocabulary size must be larger than zero")

        assert self.nmk.shape == (M, K)
        assert self.nm.shape == (M,)
        assert self.nkt.shape == (K, V)
        assert self.nk.shape == (K,)

    def _initialize_topics(self, docs, topics):
        # type: (IterableDocuments, TopicsMapping) -> None

        for m, doc in enumerate(progress(docs, file=stderr)):
            for i, t in enumerate(doc):
                k = topics[m, i] = self.init_topic()
                self.add_counts(m, t, k)

    def calc_probs(self, m, t, k_select=slice(None)):
        # type: (int, int, Indices) -> np.ndarray

        # notes: np.sum(self.nkt, axis=-1) == self.nk
        # notes: np.sum(self.nmk[m], axis=-1) == self.nm[m]

        """This is eq1. in the DC-LDA paper.

        returns: float[K]
        """

        """ The full term for left would be:
            left = (self.nmk[m, :] + self.α) / (self.nm[m] + self.αsum) # [K]
            however, the denominator can be dropped, because it doesn't depend on `k`
            and we only care about proportionality.
            It is done so in: 'A Theoretical and Practical Implementation Tutorial
                on Topic Modeling and Gibbs Sampling'
        """

        left = self.nmk[m, k_select] + self.α  # [K]
        right = (self.nkt[k_select, t] + self.β) / (self.nk[k_select] + self.βsum)  # [K]
        return left * right  # [K]

    def sample(self, m, t):
        # type: (int, int) -> int

        probs = self.calc_probs(m, t)  # [K]
        return categorical(probs)

    def _sample_all(self, docs, topics):
        # type: (IterableDocuments, TopicsMapping) -> None

        for m, doc in enumerate(progress(docs, file=stderr)):
            for i, t in enumerate(doc):
                k = topics[m, i]
                self.sub_counts(m, t, k)
                k = topics[m, i] = self.sample(m, t)  # sample new topics
                self.add_counts(m, t, k)

    def initialize(self):
        # type: () -> None

        """Set base parameters"""

        raise NotImplementedError

    def initialize_docs(self):
        # type: () -> None

        """Set parameters based on input docs"""

        raise NotImplementedError

    def initialize_topics(self):  # step1
        # type: () -> None

        """Initialize topics matrix"""

        raise NotImplementedError

    def sample_all(self):  # step2 and step3
        # type: () -> None

        """Sample new topics"""

        raise NotImplementedError

    def metric(self):
        # type: () -> float

        raise NotImplementedError

    def fit(self, n_iter=100, verbose=False):
        # type: (int, bool) -> None

        """Fit LDA"""

        assert self.docs, "No documents added to model"

        self.initialize()
        self.initialize_docs()
        self.initialize_topics()

        for i in range(n_iter):
            self.sample_all()
            if verbose:
                print(f"Step #{i}, metric: {self.metric()}")

    def theta(self):
        # type: () -> np.ndarray

        """
        returns: float[M, K]
        """

        num = self.nmk + self.α  # [M, K]
        denom = self.nm + self.αsum  # [M]
        return num / denom[:, None]  # [M, K]

    def phi(self):
        # type: () -> np.ndarray

        """
        returns: float[K, V]
        """

        num = self.nkt + self.β  # [K, V]
        denom = self.nk + self.βsum  # [K]
        return num / denom[:, None]  # [K, V]

    def _perplexity(self, docs):
        # type: (IterableDocuments, ) -> float

        θ = self.theta()
        φ = self.phi()

        num = 0
        denom = 0

        for m, doc in enumerate(docs):
            for i, t in enumerate(doc):
                num += np.log(np.inner(φ[:, t], θ[m]))
            denom += len(doc)

        return exp(-num / denom)

    def docs2topics(self):
        # type: () -> np.ndarray

        """
        returns: int[M]
        """

        return np.argmax(self.theta(), axis=-1)  # [M]

    def print_topics(self, id2word, num_words=10):
        # type: (Dict[int, str], int) -> None

        topics = top_topics(id2word, self.phi(), num_words)
        print(format_topics(topics))

    @staticmethod
    def generate(theta, phi, doc_lens):
        # type: (np.ndarray, np.ndarray, Iterable[int]) -> Tuple[dict, dict]

        """
        theta: float[K, V]
        phi: float[M, K]
        """

        z = {}  # sparse
        w = {}  # sparse

        for m, length in enumerate(doc_lens):
            for i in range(length):
                k = z[(m, i)] = categorical(theta[m])  # scalar # doc to topic
                w[(m, i)] = categorical(phi[k])  # scalar # topic to word

        return z, w

    def inferencer(self):
        # type: () -> LDAInfer

        return cast(self, LDAInfer)


class LDA(LDABase):

    """LDA using collapsed Gibbs sampling
    Based on:
    - https://en.wikipedia.org/wiki/Latent_Dirichlet_allocation
    - Latent Dirichlet Allocation (2003)
    - A Theoretical and Practical Implementation Tutorial on Topic Modeling and Gibbs Sampling (2011)
    - Integrating Out Multinomial Parameters in Latent Dirichlet Allocation and Naive Bayes for Collapsed Gibbs Sampling (2010)
    """

    def __init__(self, n_topics, alpha=0.1, beta=0.01, seed=None):
        # type: (int, float, float, Optional[int]) -> None

        LDABase.__init__(self, seed)

        self.K = n_topics  # number of topics

        self.alpha = alpha
        self.beta = beta

        self.word_encoder = BatchLabelEncoder(tokenizer="nltk")  # type: BatchLabelEncoder[str]
        self.docs = []  # type: List[LDADocument]

        self.inttype = np.int32

    def initialize(self):
        # type: () -> None

        self.V = self.word_encoder.num_labels

        self.α = self.alpha / self.K  # symmetric Dirichlet prior for document-topic distribution θ
        self.β = self.beta / self.V  # symmetric Dirichlet prior for topic-word distribution φ

        self.αsum = self.alpha
        self.βsum = self.beta

        self.nkt = np.zeros((self.K, self.V), dtype=self.inttype)  # [K, V] topic-word counts
        self.nk = np.zeros((self.K,), dtype=self.inttype)  # [K] total topic counts

    def initialize_docs(self):
        # type: () -> None

        self.M = len(self.docs)

        self.nmk = np.zeros((self.M, self.K), dtype=self.inttype)  # [M, K] document-topic counts
        self.nm = np.array(
            list(map(len, self.docs)), dtype=self.inttype
        )  # [M] total document counts, not necessary for sampling, but useful for theta and phi calculation

        # note: VariableRowMatrix(0) is slower than a dict here for some reason
        self.topics = (
            {}
        )  # type: TopicsMapping # `z`, sparse document-word topics matrix (because rows can have different lengths)

    def initialize_topics(self):
        # type: () -> None

        self._validate(self.M, self.K, self.V)

        return self._initialize_topics(self.docs, self.topics)

    def sample_all(self):
        # type: () -> None

        return self._sample_all(self.docs, self.topics)

    def metric(self):
        # type: () -> float

        return self._perplexity(self.docs)

    def perplexity(self, docs):
        # type: (IterableDocuments, ) -> float

        return self._perplexity(docs)

    def transform(self, doc_lens):
        # type: (Iterable[int], ) -> Tuple[dict, dict]

        return self.generate(self.theta(), self.phi(), doc_lens)

    def add_doc(self, doc):
        # type: (str, ) -> int

        assert isinstance(doc, str)

        words = list(self.word_encoder.fit_transform(doc))

        ret = len(self.docs)
        self.docs.append(LDADocument(words))
        return ret

    def make_doc(self, doc):
        # type: (str, ) -> LDADocument

        assert isinstance(doc, str)

        words = list(self.word_encoder.transform(doc))
        return LDADocument(words)


class LDATermWeight(LDABase):

    """LDA with term weighting
    Based on: "Term Weighting Schemes for Latent Dirichlet Allocation" (2010)

    Further analysis papers:
            - "Pulling Out the Stops: Rethinking Stopword Removal for Topic Models" (2017)
            - "Assessing topic model relevance: Evaluation and informative priors" (2019)
    """

    def __init__(self, n_topics, tws="PMI", alpha=0.1, beta=0.01):
        # type: (int, str, float, float) -> None

        """tws: Term weighting scheme
        ONE: Consider every term equal (same as standard LDA)
        TF: Term frequency
        IDF: Inverse Document Frequency term weighting.
                Thus, a term occurring at almost every document has very low weighting and a term occurring at a few document has high weighting.
        PMI: Pointwise Mutual Information term weighting (default)
        """

        LDABase.__init__(self)

        self.K = n_topics  # number of topics

        self.alpha = alpha
        self.beta = beta

        self.word_encoder = BatchLabelEncoder(tokenizer="nltk")  # type: BatchLabelEncoder[str]
        self.docs = []  # type: List[LDADocument]

        self.create_term_weights = {
            "ONE": self._one,
            "TF": self._tf,
            "IDF": self._idf,
            "TFIDF": self._tfidf,
            "PMI": self._pmi,
            "RAND": self._rand,
        }[tws]

        self.floattype = np.float32
        self.inttype = np.int32

    def _one(self, docs):
        # type: (Any, ) -> np.ndarray

        """
        returns: int[M, V]
        """

        # return np.ones((self.M, self.V))
        return np.broadcast_to(1, (self.M, self.V))  # is this slower for indexing later?

    def _tf(self, docs):
        # type: (IterableDocuments, ) -> np.ndarray

        """
        returns: float[M, V]
        """

        all_counts = Counter(chain.from_iterable(docs))
        num_counts = sum(all_counts.values())

        tw = np.array([all_counts[t] for t in range(self.V)], dtype=self.floattype)  # [V]
        tw = -np.log2(tw / num_counts)

        return np.broadcast_to(tw, (self.M, self.V))

    def _idf(self, docs):
        # type: (IterableDocuments, ) -> np.ndarray

        """
        returns: float[M, V]
        """

        tw = np.zeros((self.V,), dtype=self.floattype)  # [V]

        for doc in docs:
            for token in set(doc):
                tw[token] += 1.0

        tw = np.log2(self.M / tw)

        return np.broadcast_to(tw, (self.M, self.V))

    def _tfidf(self, docs):
        raise NotImplementedError

    def _pmi(self, docs):
        # type: (IterableDocuments, ) -> np.ndarray

        """
        returns: float[M, V]
        """

        tw = np.zeros((self.M, self.V), dtype=self.floattype)  # [M, V] term weights

        all_counts = Counter(chain.from_iterable(docs))

        for m, doc in enumerate(docs):
            doc_counts = Counter(doc)
            for i, t in enumerate(doc):
                tw[m, t] = -log2(doc_counts[t] / all_counts[t])

        np.clip(tw, 0.0, None, out=tw)
        return tw

    def _rand(self, docs):
        # type: (Any, ) -> np.ndarray

        """
        returns: float[M, V]
        """

        return np.random.uniform(0, 1, (self.M, self.V)).astype(self.floattype)

    def initialize(self):
        # type: () -> None

        self.V = self.word_encoder.num_labels

        self.α = self.alpha / self.K  # symmetric Dirichlet prior for document-topic distribution θ
        self.β = self.beta / self.V  # [V] symmetric Dirichlet prior for topic-word distribution φ

        self.αsum = self.alpha
        self.βsum = self.beta

        self.nkt = np.zeros((self.K, self.V), dtype=self.inttype)  # [K, V] counts, same as np.sum(nmkt, axis=0)
        self.nk = np.zeros((self.K,), dtype=self.inttype)  # [K] counts, same as np.sum(nkt, axis=-1) # [K]

        # self.wkt = np.zeros((self.K, self.V), dtype=self.floattype) # [K, T] weights
        # self.wk = np.zeros((self.K, ), dtype=self.floattype) # [K] weights

    def initialize_docs(self):
        # type: () -> None

        self.M = len(self.docs)

        self.tw = self.create_term_weights(self.docs)
        print(self.tw.shape)
        print(self.tw)

        self.nmk = np.zeros((self.M, self.K), dtype=self.inttype)  # [M, K] counts, same as np.sum(nmkt, axis=-1)
        self.nm = np.array(
            list(map(len, self.docs)), dtype=self.inttype
        )  # [M] total document counts, not necessary for sampling, but useful for theta and phi calculation

        """
            `wmk[m]` is same as `np.sum(wmkt[m], axis=-1)` is same as
            `np.sum(self.tw[m, :][None, :] * nmkt[m], axis=-1)`

            with
                `wmkt = np.zeros((self.M, self.K, self.V), dtype=self.floattype)` # [M, K, V] weights
                `nmkt = np.zeros((self.M, self.K, self.V), dtype=self.inttype)` # [M, K, V] counts
        """
        # self.nmkt = np.zeros((self.M, self.K, self.V), dtype=self.inttype)
        self.wmk = np.zeros((self.M, self.K), dtype=self.floattype)  # [M, K] weights, same as np.sum(nmkt, axis=-1)

        self.topics = VariableRowMatrix(
            0
        )  # type: VariableRowMatrix[int] # [M, Variable] `z`, sparse document-word topics matrix

    def initialize_topics(self):
        # type: () -> None

        # self._validate(self.M, self.K, self.V)

        self._initialize_topics(self.docs, self.topics)

    def calc_probs(self, m, t):
        # type: (int, int) -> np.ndarray

        """
        returns: float[K]
        """

        # here using global weights works
        # wmk = np.sum(self.tw[m, :][None, :] * self.nmkt[m], axis=-1, dtype=self.floattype)
        # if not np.allclose(self.wmk[m], wmk, atol=1e-06): # after some time it still diverges because of nummerical issues
        #     raise RuntimeError("inconsistent wmk")

        # here it doesn't...
        wkt = self.tw[m, t] * self.nkt[:, t]
        # if not np.allclose(self.wkt[:, t], wkt[:]):
        #     raise RuntimeError("inconsistent wkt")

        # here it doesn't also...
        wk = np.sum(self.tw[m, :][None, :] * self.nkt, axis=-1)  # not the same as `self.wk[k] (+|-)= self.tw[m, t]`
        # if not np.allclose(self.wk, wk):
        #     raise RuntimeError("inconsistent wk")

        # nk: ld.numByTopic[tid]
        # nkt: ld.numByTopicWord[tid, vid]

        left = self.wmk[m] + self.α  # [K]
        right = (wkt[:] + self.β) / (wk[:] + self.βsum)  # [K]
        return left * right  # [K]

    """
    def addWordToINC(docId, tid, vid):

        # paper W*beta term in eq. 6 (https://www.aclweb.org/anthology/N10-1070.pdf)
        ld.numByTopicWordCOUNTS[tid, vid] += 1
        Nk = np.sum(docs[docId].wordWeights[None,:] * ld.numByTopicWordCOUNTS, axis=-1) # [None,:] needed for correct broadcasting. summed over words, K size array remains
        # or the above in plain python
        Nk = [sum(docs[docId].wordWeights[word_id] * ld.numByTopicWordCOUNTS[tid, word_id]
            for word_id in docs[docid].words)
                for tid in range(K)] # K size array

        # addWordTo INC
        ld.numByTopic[tid] += docs[docId].wordWeights[pid] # ld.numByTopic is a K size array

        # now both weight sums should be the same AT ALL TIMES
        assert Nk == ld.numByTopic # elementwise comparison

        # but they are not after the first document, if the weights depend on the document
    """

    def add_counts(self, m, t, k):
        # type: (int, int, int) -> None

        self.nmk[m, k] += 1
        self.nkt[k, t] += 1
        self.nk[k] += 1

        # self.nmkt[m, k, t] += 1
        self.wmk[m, k] += self.tw[m, t]
        # self.wkt[k, t] += self.tw[m, t]
        # self.wk[k] += self.tw[m, t]

        """
        print("docId={}, vid={}, tid={}".format(m, t, k))

        wk = np.sum(self.tw[m, :][None, :] * self.nkt, axis=-1)
        print("ld.numByTopicWordCOUNTS:\n", self.nkt)
        print("paper sum:", wk)
        print("ld.numByTopic:", self.wk)

        if not np.allclose(wk, self.wk):
            raise RuntimeError("not equal")
        """

    def sub_counts(self, m, t, k):
        # type: (int, int, int) -> None

        self.nmk[m, k] -= 1
        self.nkt[k, t] -= 1
        self.nk[k] -= 1

        # self.nmkt[m, k, t] -= 1
        self.wmk[m, k] -= self.tw[m, t]
        # self.wkt[k, t] -= self.tw[m, t]
        # self.wk[k] -= self.tw[m, t]

    def sample_all(self):
        # type: () -> None

        self._sample_all(self.docs, self.topics)

    def metric(self):
        # type: () -> float

        return self._perplexity(self.docs)


class LDAInfer(LDA):
    def __init__(self):
        # type: () -> None

        self.nkt_old = self.nkt
        self.nk_old = self.nk
        self.docs = []

    def calc_probs(self, m, t, k_select=slice(None)):
        # type: (int, int, Indices) -> np.ndarray

        """
        returns: float[K]
        """

        left = self.nmk[m, k_select] + self.α  # [K]
        num = self.nkt[k_select, t] + self.nkt_old[k_select, t] + self.β  # [K]
        denom = self.nk[k_select] + self.nk_old[k_select] + self.βsum  # [K]

        return left * num / denom  # [K]

    def initialize(self):
        # type: () -> None

        assert self.K > 0, "Number of topics must be larger than zero"
        assert self.V > 0, "Vocabulary size must be larger than zero"

        self.nkt = np.zeros((self.K, self.V), dtype=self.inttype)
        self.nk = np.zeros((self.K,), dtype=self.inttype)


def test_lda_20():
    from sklearn.datasets import fetch_20newsgroups

    ng = fetch_20newsgroups()

    lda = LDA(n_topics=20, seed=0)

    for doc in ng.data[:1000]:
        lda.add_doc(doc)

    lda.fit(10, verbose=True)

    return lda


def test_ldatw_20():
    from sklearn.datasets import fetch_20newsgroups

    ng = fetch_20newsgroups()

    lda = LDATermWeight(n_topics=20)

    for doc in ng.data[:1000]:
        lda.add_doc(doc)

    lda.fit(10, verbose=True)

    return lda


if __name__ == "__main__":
    from genutility.time import PrintStatementTime

    for i in range(3):
        with PrintStatementTime():
            test_lda_20()
            # test_ldatw_20()



from __future__ import generator_stop

from typing import TYPE_CHECKING

from ruamel.yaml import YAML

if TYPE_CHECKING:
    from typing import Any


def read_yaml(path):
    # type: (str, ) -> Any

    with open(path, encoding="utf-8") as fr:
        yaml = YAML(typ="safe")
        return yaml.load(fr)


def write_yaml(obj, path):
    # type: (str, Any) -> None

    with open(path, "w", encoding="utf-8") as fw:
        yaml = YAML(typ="safe")
        yaml.dump(obj, fw)



from __future__ import generator_stop

from collections import defaultdict
from datetime import timedelta
from time import monotonic, sleep
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, DefaultDict, Dict, Hashable, Iterator, List, Optional, Tuple, TypeVar, Union

    T = TypeVar("T")


def iter_timer(it):
    # type: (Iterator[T], ) -> Iterator[Tuple[T, float]]
    try:
        while True:
            start = monotonic()
            res = next(it)
            delta = monotonic() - start
            yield res, delta
    except StopIteration:
        pass


class TakeAtleast:

    __slots__ = ("delta", "wait_on_error", "now")

    def __init__(self, delta, wait_on_error=False):
        # type: (Union[float, timedelta], bool) -> None

        if isinstance(delta, timedelta):
            self.delta = delta.total_seconds()
        else:
            self.delta = float(delta)

        self.wait_on_error = wait_on_error

    def __enter__(self):
        self.now = monotonic()

    def __exit__(self, exc_type, exc_value, traceback):

        if exc_type is None or self.wait_on_error:
            elapsed = monotonic() - self.now
            if elapsed < self.delta:
                sleep(self.delta - elapsed)


class DeltaTime:

    __slots__ = ("start", "end")

    def __init__(self):
        # type: () -> None

        self.start = None  # type: Optional[float]
        self.end = monotonic()

    def __iter__(self):
        # type: () -> DeltaTime

        return self

    def __next__(self):
        # type: () -> float

        self.start, self.end = self.end, monotonic()
        return self.end - self.start

    def get(self):
        # type: () -> float

        return monotonic() - self.end

    def reset(self):
        # type: () -> None

        self.end = monotonic()

    def get_reset(self):
        # type: () -> float

        self.start, self.end = self.end, monotonic()
        return self.end - self.start


class PrintStatementTime:

    __slots__ = ("tpl", "start")

    def __init__(self, tpl=None):
        # type: (Optional[str], ) -> None
        if tpl is None:
            self.tpl = "Execution took {delta}s"
        else:
            self.tpl = tpl

    def __enter__(self):
        self.start = monotonic()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        end = monotonic()
        delta = end - self.start
        msg = self.tpl.format(start=self.start, end=end, delta=delta)
        if exc_type:
            print(msg + " (interrupted)")
        else:
            print(msg)


class MeasureTime:

    __slots__ = ("delta", "start")

    def __init__(self):
        # type: () -> None

        self.delta = None  # type: Optional[float]

    def __enter__(self):
        # type: () -> MeasureTime

        self.start = monotonic()
        return self

    def __exit__(self, type, value, traceback):
        self.delta = monotonic() - self.start

    def get(self):
        # type: () -> float

        if self.delta:
            return self.delta
        else:
            return monotonic() - self.start


class TimeIt:
    def __init__(self):
        # type: () -> None

        self.results = defaultdict(list)  # type: DefaultDict[Hashable, List[float]]
        self.starts = dict()  # type: Dict[Hashable, DeltaTime]

    def __call__(self, key, func, *args, **kwargs):
        # type: (Hashable, Callable[..., T], *Any, **Any) -> T

        with MeasureTime() as t:
            ret = func(*args, **kwargs)
        self.results[key].append(t.get())
        return ret

    def start(self, key):
        # type: (Hashable, ) -> None

        self.starts[key] = DeltaTime()

    def stop(self, key):
        # type: (Hashable, ) -> None

        self.results[key].append(self.starts[key].get())

    def min(self, key):
        # type: (Hashable, ) -> float

        return min(self.results[key])

    def length(self, key):
        # type: (Hashable, ) -> int

        return len(self.results[key])



from __future__ import generator_stop

from itertools import chain, repeat
from typing import TYPE_CHECKING

from .exceptions import InconsistentState, NoResult

if TYPE_CHECKING:
    from typing import Any, Iterator

    from .typing import Connection, Cursor


class TransactionCursor:

    """Cursor context manager which starts a transaction and rolls back in case of error."""

    def __init__(self, conn):
        # type: (Connection, ) -> None

        self.cursor = conn.cursor()

    def __enter__(self):
        # type: () -> Cursor

        self.cursor.execute("BEGIN TRANSACTION")
        return self.cursor

    def __exit__(self, exc_type, exc_value, traceback):
        # type: (Any, Any, Any) -> None

        if exc_type:
            self.cursor.execute("ROLLBACK TRANSACTION")
        else:
            self.cursor.execute("COMMIT TRANSACTION")
        self.cursor.close()


class CursorContext:

    """Cursor context manager which creates a new cursor and closes it when it leaves the context."""

    def __init__(self, conn):
        # type: (Connection, ) -> None

        self.cursor = conn.cursor()

    def __enter__(self):
        # type: () -> Cursor

        return self.cursor

    def __exit__(self, exc_type, exc_value, traceback):
        # type: (Any, Any, Any) -> None

        self.cursor.close()


def upsert(cursor, primary, values, table):
    # type: (Cursor, dict, dict, str) -> bool

    """Inserts `table` fields specified by `primary` and `values` keys,
    with the corresponding values.
    If all the values specified by `primary` already exist,
    they are updated instead.

    `table`, `primary.keys()` and `values.keys()` are not escaped.
    `primary.values()` and `values.values()` are escaped.
    """

    # use INSERT ... ON DUPLICATE KEY UPDATE instead?

    if not primary:
        raise ValueError("Empty primary mapping would result in an empty WHERE condition which would affect all rows")

    set_str = ",".join(f"{k}=?" for k in values.keys())
    where_str = " AND ".join(f"{k}=?" for k in primary.keys())

    cursor.execute(f"UPDATE {table} SET {set_str} WHERE {where_str}", chain(values.values(), primary.values()))  # nosec

    if cursor.rowcount == 0:
        into_str = ",".join(chain(primary.keys(), values.keys()))
        values_str = ",".join(repeat("?", len(primary) + len(values)))
        cursor.execute(
            f"INSERT INTO {table} ({into_str}) VALUES ({values_str})", chain(primary.values(), values.values())  # nosec
        )
        return True

    return False


def fetchone(cursor):
    # type: (Cursor, ) -> Any

    """Fetch results from `cursor` and assure only one result was returned."""

    rows = cursor.fetchall()
    if len(rows) == 0:
        raise NoResult("No result found")
    elif len(rows) == 1:
        return rows[0]
    else:
        raise InconsistentState("More than one result found")


def iterfetch(cursor, batchsize=1000):
    # type: (Cursor, int) -> Iterator[Any]

    """Iterate all results from `cursor`."""

    while True:
        results = cursor.fetchmany(batchsize)
        if not results:
            break
        yield from results



from __future__ import generator_stop

import os.path
from os import DirEntry
from typing import Union

""" This module is to avoid circular imports.
    It should avoid any dependencies apart from the standard library.
"""


class BaseDirEntry:
    __slots__ = ("entry",)

    def __init__(self, entry):
        self.entry = entry

    @property
    def name(self):
        return self.entry.name

    @property
    def path(self):
        return self.entry.path

    def inode(self):
        return self.entry.inode()

    def is_dir(self):
        return self.entry.is_dir()

    def is_file(self):
        return self.entry.is_file()

    def is_symlink(self):
        return self.entry.is_symlink()

    def stat(self):
        return self.entry.stat()

    def __str__(self):
        return str(self.entry)

    def __repr__(self):
        return repr(self.entry)

    def __fspath__(self):
        return self.entry.path


MyDirEntryT = Union[DirEntry, BaseDirEntry]


def entrysuffix(entry):
    # type: (MyDirEntryT, ) -> str

    return os.path.splitext(entry.name)[1]



from __future__ import generator_stop

import tracemalloc
from typing import Optional


class MeasureMemory:

    __slots__ = ("total", "snapshot")

    def __init__(self):
        # type: () -> None

        tracemalloc.start()

        self.total = None  # type: Optional[int]

    def _comp(self):
        # type: () -> int

        snapshot_now = tracemalloc.take_snapshot()
        stats = snapshot_now.compare_to(self.snapshot, "lineno")
        return sum(stat.size for stat in stats)

    def __enter__(self):
        # type: () -> MeasureMemory

        self.snapshot = tracemalloc.take_snapshot()
        return self

    def __exit__(self, type, value, traceback):
        self.total = self._comp()

    def get(self):
        # type: () -> int

        if self.total:
            return self.total
        else:
            return self._comp()

    def print(self, name):
        # type: (str, ) -> None

        if self.total:
            total = self.total
        else:
            total = self._comp()

        print(f"{name} uses {total / 1024 / 1024:.3f} MiB of memory")



from __future__ import generator_stop

import logging
import os.path
from datetime import datetime, timedelta
from functools import partial, reduce, wraps
from sys import stdout
from time import sleep
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from ._func import rename, renameobj  # noqa: F401
from .iter import nonstriter, retrier

if TYPE_CHECKING:
    T = TypeVar("T")
    U = TypeVar("U")

logger = logging.getLogger(__name__)


class NotRetried(RuntimeError):
    pass


def identity(x):
    # type: (T, ) -> T

    """Identity function."""

    return x


def nop():
    # type: () -> None

    """Function which does absolutely nothing (aka pass, noop)."""

    pass


def partial_decorator(*args, **kwargs):
    # type: (*Any, **Any) -> Callable

    """Same as `functools.partial` but applied as a decorator."""

    def decorator(func):
        return partial(func, *args, **kwargs)

    return decorator


def compose_two(f, g):
    # type: (Callable[[Any], Any], Callable[[Any], Any]) -> Callable[[Any], Any]

    """compose_two(f, g) -> lambda x: f(g(x))"""

    return lambda x: f(g(x))


def compose(*functions):
    # type: (*Callable[[Any], Any]) -> Callable[[Any], Any]

    """compose(f, g, h) -> lambda x: f(g(h(x))).
    see: Function composition
    """

    return reduce(compose_two, functions, identity)


def apply(f, x):  # *args, **kwargs
    # type: (Callable[[T], U], T) -> U

    """Same as `f(x)`."""

    return f(x)


def zipmap(funcs, vals):
    # type: (Iterable[Callable[[T], U]], Iterable[T]) -> Iterator[U]

    """Applies a list of functions to a list of values."""

    return map(apply, funcs, vals)


def multiapply(funcs, elm):
    # type: (Iterable[Callable], Any) -> Any

    """Applies functions `funcs` to element `elm` iteratively."""

    for func in funcs:
        elm = func(elm)

    return elm


def multimap(funcs, it):
    # type: (Sequence[Callable], Iterable[Any]) -> Iterator[Any]

    """Applies functions `funcs` to each element of `it` iteratively."""

    for i in it:
        yield multiapply(funcs, i)


def deepmap(func, *iterables):
    # type: (Callable, *Iterable) -> Iterator

    """Simlar to `map`, but it maps recursively over sequences of sequnces.
    Returns a generator of generators.
    To output plain objects use `recmap` instead.

    Example:
    deepmap(lambda x: x*2, [1, [2, 3]]) -> (2, (4, 6))
    """

    if not iterables:
        raise ValueError("iterables must have a least one element")

    for args in zip(*iterables):
        try:
            its = tuple(map(nonstriter, args))
        except TypeError:
            yield func(*args)
        else:
            yield deepmap(func, *its)


def outermap(func, iterable):
    # type: (Callable[[T], U], Iterable) -> U

    """Examples:
    outermap(list, (1, (2, 3))) -> [1, [2, 3]]
    outermap(sum, (1, (2, 3)) -> 6
    """

    try:
        it = nonstriter(iterable)
    except TypeError:
        return iterable
    else:
        return func(outermap(func, i) for i in it)


def recmap(func, iterable):
    # type: (Callable, Iterable) -> list

    """Simlar to `map`, but it maps recursively over sequences of sequnces and puts the result into a list of lists.
    To return a generator of generators use `deepmap` instead.

    Example:
    recmap(lambda x: x*2, [1, [2, 3]]) -> [2, [4, 6]]
    """

    return outermap(list, deepmap(func, iterable))


def call_repeated(num):
    # type: (int, ) -> Callable[[Callable], Callable]

    """Function decorator to call decorated function `num` times with the same arguments.
    Returns the results of the last call.
    """

    if num < 1:
        raise ValueError("num must be larger than 0")

    def dec(func):
        @wraps(func)
        def inner(*args, **kwargs):
            last = None
            for i in range(num):
                last = func(*args, **kwargs)
            return last

        return inner

    return dec


def print_return_type(func, file=stdout):
    # type: (Callable, TextIO) -> Callable

    """Wraps function to print the return type after calling."""

    @wraps(func)
    def inner(*args, **kwargs):
        ret = func(*args, **kwargs)
        print(type(ret), file=file)
        return ret

    return inner


def retry(
    func,
    waittime,
    exceptions=(Exception,),
    attempts=-1,
    multiplier=1,
    jitter=0,
    max_wait=None,
    jitter_dist="uniform",
    waitfunc=sleep,
):
    # type: (Callable[[], T], float, Union[Type[Exception], Tuple[Type[Exception]]], int, float, float, Optional[float], str, Callable[[float], Any]) -> T

    """Retry function `func` multiple times in case of raised `exceptions`.
    See `genutility.iter.retrier()` for the remaining arguments.
    Reraises the last exception in case the function call doesn't succeed after retrying.
    """

    last_exception = None  # type: Optional[Exception]
    for i in retrier(waittime, attempts, multiplier, jitter, max_wait, jitter_dist, waitfunc):
        try:
            return func()
        except exceptions as e:
            logger.info("Attempt %s failed: %s", i + 1, e)
            last_exception = e

    if last_exception:
        raise last_exception  # pylint: disable=raising-bad-type
    else:
        raise NotRetried


class CustomCache:

    """Class to build decorator cache function using custom reader and writer functions.
    The cache method ignores all arguments of the decorated function.

    Example:
    ```
    cc = CustomCache(read_pickle, write_pickle)

    @cc.cache("func-cache.p")
    # arg is ignored, so the cache will return the result of the argument which was supplied
    # when the cache file was created.
    def func(arg):
            return arg
    ```
    a = func(1) # cache created, a == 1
    b = func(2) # cache loaded, b == 1
    """

    def __init__(self, reader, writer):
        # type: (Callable, Callable) -> None

        self.reader = reader
        self.writer = writer

    def cache(self, path):
        # type: (str, ) -> Callable

        def dec(func):
            # type: (Callable, ) -> Callable

            @wraps(func)
            def inner(*args, **kwargs):
                # type: (*Any, **Any) -> Any

                if os.path.exists(path):
                    logger.debug("Loading object from cache %s", path)
                    return self.reader(path)
                else:
                    logger.debug("Saving obect to cache %s", path)
                    result = func(*args, **kwargs)
                    self.writer(result, path)
                    return result

            return inner

        return dec


class RunScheduled:
    def __init__(self, delta: timedelta, func: Callable):
        self.delta = delta
        self.func = func
        self.lastrun: Optional[datetime] = None

    def __call__(self, *args, **kwargs):
        now = datetime.now()
        if self.lastrun is None or now - self.lastrun > self.delta:
            self.func(*args, **kwargs)
            self.lastrun = now



from __future__ import generator_stop

from logging import Formatter
from typing import TYPE_CHECKING

from .datetime import datetime_from_utc_timestamp

if TYPE_CHECKING:
    from logging import LogRecord
    from typing import Any, Dict, Optional


class IsoDatetimeFormatter(Formatter):

    """Displays the time in ISO 8601 format.
    Instead of passing a formatting string to `datefmt` in the initializer, `sep` and `timespec`
    (see `datetime.datetime.isoformat`) as well as `aslocal` (see genutility.datetime.datetime_from_utc_timestamp`)
    can be passed.
    """

    def __init__(self, fmt=None, datefmt=None, style="%", validate=True, sep="T", timespec="auto", aslocal=False):
        # type: (Optional[str], None, str, bool, str, str, bool) -> None

        Formatter.__init__(self, fmt, datefmt, style, validate)
        assert datefmt is None
        self.sep = sep
        self.timespec = timespec
        self.aslocal = aslocal

    def formatTime(self, record, datefmt):
        # type: (LogRecord, None) -> str

        return datetime_from_utc_timestamp(record.created, aslocal=self.aslocal).isoformat(self.sep, self.timespec)


class OverwriteFormatter(Formatter):

    """Usually the default Formatter disallows overwriting of built-in fields
    like filename, funcName, lineno and so on. Here you can specify a mapping
    to set these values anyway.

    Example:
            formatter = OverwriteFormatter({"func_name": "funcName"}, "%(funcName)s: %(message)s")
            and then pass `extra={"func_name": "thename"}` to the logging calls.
            If the the custom fields are not passed, the original values will be used.
    """

    def __init__(self, map, *args, **kwargs):
        # type: (Dict[str, str], *Any, **Any) -> None

        self.map = map
        Formatter.__init__(self, *args, **kwargs)

    def format(self, record):
        # type: (LogRecord, ) -> str

        for k, v in self.map.items():
            try:
                setattr(record, v, getattr(record, k))
            except AttributeError:
                pass

        return Formatter.format(self, record)


if __name__ == "__main__":
    import logging

    logger = logging.getLogger(__name__)
    handler = logging.StreamHandler()
    formatter = OverwriteFormatter({"func_name": "funcName"}, "%(funcName)s: %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    logger.warning("asd")
    logger.warning("asd", extra={"func_name": "thename"})



from __future__ import generator_stop

import logging
import os.path
import sqlite3
from itertools import chain
from os import fspath
from typing import TYPE_CHECKING

from tls_property import tls_property

from .exceptions import NoResult
from .filesystem import normalize_seps
from .sql import fetchone, iterfetch
from .sqlite import quote_identifier

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from typing import Any, Dict, FrozenSet, Iterable, Iterator, Optional, Tuple

    from .filesystem import EntryType


class GenericFileDb:
    def __init__(self, dbpath, table, debug=True):
        # type: (str, str, bool) -> None

        self.table = quote_identifier(table)
        self._primary = self.primary()
        self._auto = self.auto()
        self._mandatory = self.mandatory()
        self._derived = self.derived()

        self.dbpath = dbpath
        self.setup()

        if logger.isEnabledFor(logging.DEBUG):
            logger.info("SQLite tracing enabled")
            self.connection.set_trace_callback(self.trace)

    @tls_property
    def connection(self):
        return sqlite3.connect(self.dbpath, check_same_thread=False)

    @tls_property
    def cursor(self):
        return self.connection.cursor()

    def trace(self, query):
        logger.debug("SQL trace: %s", query)

    @classmethod
    def latest_order_by(cls):
        return "entry_date DESC"

    @classmethod
    def primary(cls):
        """Never explicitly specified.
        Not needed if there is a PRIMARY KEY in one of the other fields.
        """

        raise NotImplementedError

    @classmethod
    def auto(cls):
        """Explicitly specified value.
        Can never be a ? placeholder, but must be a SQL value like `datetime('now')`.
        """

        raise NotImplementedError

    @classmethod
    def mandatory(cls):
        """Mandatory fields used to retrieve rows.
        Used for all search queries.
        """

        raise NotImplementedError

    @classmethod
    def derived(cls):
        """Optional fields used to retrieve rows.
        Used for search queries when given.
        """

        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        # type: () -> None

        """Only closes connections and cursors opened in the current thread"""

        self.cursor.close()
        self.connection.close()

    def setup(self):
        # type: () -> None

        fields = ", ".join(f"{n} {t}" for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived))

        # unique = "UNIQUE({})".format(n for n, t, v in chain(self._mandatory, self._derived))
        # add to sql below?
        sql = f"CREATE TABLE IF NOT EXISTS {self.table} ({fields})"

        self.cursor.execute(sql)
        self.commit()

    def normalize_path(self, path):
        raise NotImplementedError

    def commit(self):
        # type: () -> None

        self.connection.commit()

    def _args(self, path, filesize, mod_date, derived=None, ignore_null=True):
        # type: (str, int, int, Optional[Dict[str, Any]], bool) -> tuple

        derived = derived or {}
        if ignore_null:
            return tuple(
                chain(
                    (self.normalize_path(path), filesize, mod_date),
                    (derived[n] for n, t, v in self._derived if n in derived),
                )
            )
        else:
            return tuple(
                chain((self.normalize_path(path), filesize, mod_date), (derived.get(n) for n, t, v in self._derived))
            )

    def iter(self, only=frozenset(), no=frozenset()):
        # type: (FrozenSet[str], FrozenSet[str]) -> Iterator[tuple]

        if only and no:
            raise ValueError("Only `only` or `no` can be specified")

        if only:
            fields = ", ".join(
                n for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived) if n in only
            )
        else:
            fields = ", ".join(
                n for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived) if n not in no
            )

        self.cursor.execute(f"SELECT {fields} FROM {self.table}")  # nosec
        return iterfetch(self.cursor)

    def get_latest(self, path, filesize, mod_date, derived=None, ignore_null=True, only=frozenset(), no=frozenset()):
        # type: (str, int, int, Optional[Dict[str, Any]], bool, FrozenSet[str], FrozenSet[str]) -> tuple

        """Retrieve latest row based on mandatory and derived information.

        If `ignore_null=True` (default), values which are not provided in `derived`
        are ignored for the search.
        If it's `False`, values not provided in `derived` are assumed to be NULL
        and the row is only matches if the values in the database are NULL as well.

        The `only` argument can be used to include only these fields fields in the output.
        The `no` argument can be used to omit fields from the output.
        Only one of them can be specified.
        """

        if only and no:
            raise ValueError("Only `only` or `no` can be specified")

        derived = derived or {}
        args = self._args(path, filesize, mod_date, derived, ignore_null)

        if ignore_null:
            _derived = ((n, t, v) for n, t, v in self._derived if n in derived)
        else:
            _derived = self._derived

        if only:
            fields = ", ".join(
                n for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived) if n in only
            )
        else:
            fields = ", ".join(
                n for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived) if n not in no
            )

        if not fields:
            raise ValueError("No output fields selected")

        conditions = " AND ".join(f"{n} IS ?" for n, t, v in chain(self._mandatory, _derived))
        order_by = self.latest_order_by()
        sql = f"SELECT {fields} FROM {self.table} WHERE {conditions} ORDER BY {order_by} LIMIT 1"  # nosec

        self.cursor.execute(sql, args)

        return fetchone(self.cursor)

    def get(self, path, only=frozenset(), no=frozenset()):
        # type: (EntryType, FrozenSet[str], FrozenSet[str]) -> tuple

        """Retrieves latest row based on mandatory information
        which is solely based on the `path`.
        Use `only`/`no` to include/exclude returned fields.
        """

        stats = path.stat()
        return self.get_latest(fspath(path), stats.st_size, stats.st_mtime_ns, ignore_null=True, only=only, no=no)

    def _add_file(self, path, filesize, mod_date, derived=None):
        # type: (str, int, int, Optional[Dict[str, Any]]) -> None

        """Adds a new entry to the database and doesn't check if file
        with the same mandatory fields already exists.
        However it will replace entries based on PRIMARY KEYs or UNIQUE indices
        """

        args = self._args(path, filesize, mod_date, derived, ignore_null=False)

        fields = ", ".join(n for n, t, v in chain(self._auto, self._mandatory, self._derived))
        values = ", ".join(v for n, t, v in chain(self._auto, self._mandatory, self._derived))
        sql = f"REPLACE INTO {self.table} ({fields}) VALUES ({values})"
        self.cursor.execute(sql, args)

    def _add_file_no_dup(self, path, filesize, mod_date, derived=None, ignore_null=True):
        # type: (str, int, int, Optional[Dict[str, Any]], bool) -> bool

        """Only adds a new entry to the db if the provided information
        does not exist in the db yet.
        """

        try:
            self.get_latest(path, filesize, mod_date, derived, ignore_null=ignore_null)
            return False

        except NoResult:
            self._add_file(path, filesize, mod_date, derived)
            return True

    def add(self, path, derived=None, commit=True):
        # type: (EntryType, Optional[Dict[str, Any]], bool) -> None

        stats = path.stat()
        self._add_file(fspath(path), stats.st_size, stats.st_mtime_ns, derived)
        if commit:
            self.commit()

    def add_file(self, path, filesize, mod_date, derived=None):
        # type: (str, int, int, Optional[Dict[str, Any]]) -> bool

        result = self._add_file_no_dup(path, filesize, mod_date, derived)
        self.commit()
        return result

    def add_files(self, batch):
        # type: (Iterable[Tuple[str, int, int, Optional[Dict[str, Any]]]], ) -> Iterator[bool]

        for path, filesize, mod_date, derived in batch:
            yield self._add_file_no_dup(path, filesize, mod_date, derived)

        self.commit()


class FileDbHistory(GenericFileDb):
    @classmethod
    def primary(cls):
        return [
            ("file_id", "INTEGER NOT NULL PRIMARY KEY", "?"),
        ]

    @classmethod
    def auto(cls):
        return [
            ("entry_date", "DATETIME", "datetime('now')"),
        ]

    @classmethod
    def mandatory(cls):
        return [
            ("path", "VARCHAR(256)", "?"),
            ("filesize", "INTEGER", "?"),
            ("mod_date", "INTEGER", "?"),
        ]

    def normalize_path(self, path):
        drive, path = os.path.splitdrive(path)
        return normalize_seps(os.path.normpath(path))


class FileDbSimple(GenericFileDb):
    @classmethod
    def primary(cls):
        return []

    @classmethod
    def auto(cls):
        return [
            ("entry_date", "DATETIME", "datetime('now')"),
        ]

    @classmethod
    def mandatory(cls):
        return [
            ("path", "VARCHAR(256) NOT NULL PRIMARY KEY", "?"),
            ("filesize", "INTEGER", "?"),
            ("mod_date", "INTEGER", "?"),
        ]

    def normalize_path(self, path):
        return path



from __future__ import generator_stop

import csv
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, Iterator


def parse_imdb_csv(path):
    # type: (str, ) -> Iterator[Dict[str, str]]

    with open(path, encoding="iso-8859-1", errors="strict", newline="") as fr:
        csvreader = csv.DictReader(fr)
        yield from csvreader



import sys
from traceback import format_exception

from rich.console import Console
from rich.markdown import Markdown


def install_markdown_excepthook(console=None):
    console_ = console or Console(file=sys.stderr)

    def excepthook(exc_type, exc_value, exc_traceback):
        lines = list(format_exception(exc_type, exc_value, exc_traceback))
        for line in lines[:-1]:
            console_.print(line, end="")

        args = [Markdown(arg) for arg in exc_value.args]
        if args:
            console_.print(f"{exc_type.__qualname__}:", *args)
        else:
            console_.print(f"{exc_type.__qualname__}")

    old_excepthook = sys.excepthook
    sys.excepthook = excepthook
    return old_excepthook


if __name__ == "__main__":

    install_markdown_excepthook()

    MARKDOWN = """
# This is an h1

Rich can do a pretty *decent* job of rendering markdown.

1. This is a list item
2. This is another list item
"""

    raise ValueError(MARKDOWN)



from __future__ import generator_stop

import re

_markdown_newline_pat = re.compile(r"[^\n]\n[^\n]")


def _markdown_newline_func(x):
    return x.group(0).replace("\n", "  \n")


def markdown_newline(s):
    # type: (str, ) -> str

    """Convert newlines to markdown linebreaks."""

    return _markdown_newline_pat.sub(_markdown_newline_func, s)


_markdown_urls_pat = re.compile(
    r"(?<![\<])((?:http|https|ftp):\/\/[a-zA-Z0-9\-\.]+\.[a-zA-Z0-9]{2,}(?:\/[\:\/\?\#\[\]\@\!\$\&\'\(\)\*\+\;\=a-zA-Z0-9\-\._\~\%]*)?)(?![\>])"
)


def markdown_urls(s, ignore_trailing_dot=True):
    # type: (str, bool) -> str

    """Converts text with URLs to text with markdown formatted URLs.
    markdown_urls("Visit https://google.com") -> "Visit <https://google.com>"
    This function is idempotent, which means that it will not convert URLs which are already in the correct format.
    markdown_urls("Visit <https://google.com>") -> "Visit <https://google.com>"
    """

    def markdown_dotfix(m):
        (url,) = m.groups()
        if ignore_trailing_dot and url.endswith("."):
            return f"<{url[:-1]}>."
        else:
            return f"<{url}>"

    # return _markdown_urls_pat.sub(r"<\1>", s)
    return _markdown_urls_pat.sub(markdown_dotfix, s)



from __future__ import generator_stop

import pyspark.sql.functions as f
from pyspark.sql.column import Column


def timestamp_from_unixtime_ms(colname: str) -> Column:
    """Probably requires Spark>=3.0.0"""

    dt = f.from_unixtime(f.col(colname) / 1000).cast("timestamp")
    msec = f.expr("INTERVAL 1 milliseconds") * (f.col(colname) % 1000)
    return (dt + msec).alias(colname)



from __future__ import generator_stop

from typing import TYPE_CHECKING

import torch
from torch.distributions.categorical import Categorical

if TYPE_CHECKING:
    from torch import tensor


def categorical(probs, size=torch.Size()):
    # type: (tensor, tensor) -> tensor

    return Categorical(probs).sample(size)



from __future__ import generator_stop

from typing import Callable, List, Optional


class SegmentTree:

    # ported from https://codeforces.com/blog/entry/18051

    t: List[Optional[int]]

    def __init__(self, arr, func, initializer):
        # type: (List[int], Callable[[int, int], int], int) -> None

        self.n = len(arr)
        self.func = func
        self.initializer = initializer

        self.t = [None] * self.n + arr  # type

    def build(self):
        # type: () -> None

        i = self.n - 1
        while i > 0:
            self.t[i] = self.func(self.t[i << 1], self.t[i << 1 | 1])
            i -= 1

    def modify(self, p, value):
        # type: (int, int) -> None

        p += self.n
        self.t[p] = value

        while p > 1:
            val = self.func(self.t[p], self.t[p ^ 1])
            p >>= 1
            self.t[p] = val

    def query(self, left, right):
        # type: (int, int) -> int

        if left >= right or not 0 <= left <= self.n or not 0 <= right <= self.n:
            raise ValueError(f"Interval [{left}, {right}) out of range")

        res = self.initializer
        left += self.n
        right += self.n

        while left < right:

            if left & 1:
                res = self.func(res, self.t[left])
                left += 1
            if right & 1:
                right -= 1
                res = self.func(res, self.t[right])

            left >>= 1
            right >>= 1

        return res



from __future__ import generator_stop

import csv
from functools import partial
from itertools import islice
from typing import Callable, Iterator, List, Optional, Sequence

from .dict import itemgetter
from .func import compose, identity, zipmap


def iter_csv(path, delimiter=",", encoding="utf-8", skip=0):
    # type: (str, str, str, bool) -> Iterator[List[str]]

    with open(path, encoding="utf-8", newline="") as fr:
        yield from islice(csv.reader(fr, delimiter=delimiter), skip, None)


def read_csv(path, delimiter=",", skip=0, usecols=None, dtype=None, encoding="utf-8"):
    # type: (str, str, int, Optional[Sequence[int]], Optional[Sequence[Callable]], str) -> Iterator[tuple]

    with open(path, encoding=encoding, newline="") as csvfile:
        if usecols:
            if dtype:
                getcols = compose(partial(zipmap, dtype), itemgetter(usecols))
            else:
                getcols = itemgetter(usecols)
        else:
            if dtype:
                getcols = partial(zipmap, dtype)
            else:
                getcols = identity

        for row in islice(csv.reader(csvfile, delimiter=delimiter), skip, None):
            print(row)
            yield tuple(getcols(row))



from __future__ import generator_stop

from typing import TYPE_CHECKING

import piexif
from PIL import Image, ImageDraw, ImageFont
from PIL.ExifTags import GPSTAGS, TAGS

from .exceptions import NoActionNeeded

if TYPE_CHECKING:
    from typing import Any, Dict, Tuple, Union

    Color = Union[str, Tuple[int, int, int]]


def multiline_textsize(text, ttf, spacing=4):
    # type: (str, ImageFont, int) -> Tuple[int, int]

    lines = text.splitlines()

    width = 0
    height = (len(lines) - 1) * spacing

    for line in lines:
        w, h = ttf.getsize(line)
        width = max(width, w)
        height = height + h

    return width, height


def exifinfo(image):
    # type: (Image, ) -> Dict[str, Any]

    ret = dict()  # type: Dict[str, Any]

    exifd = image._getexif()

    if exifd is None:
        return dict()

    for k, v in exifd.items():
        try:
            tag = TAGS[k]
            if tag == "GPSInfo":
                ret[tag] = dict()
                for k, v in v.items():
                    try:
                        gpstag = GPSTAGS[k]
                        ret[tag][gpstag] = v
                    except KeyError:
                        pass
            else:
                ret[tag] = v
        except KeyError:
            pass

    return ret


def text_with_outline(draw, pos, text, font, fillcolor, outlinecolor, outlinesize=1):
    # type: (ImageDraw, Tuple[int, int], str, ImageFont, Color, Color, int) -> None

    x, y = pos
    delta = outlinesize

    # thin border
    draw.text((x - delta, y), text, font=font, fill=outlinecolor)
    draw.text((x + delta, y), text, font=font, fill=outlinecolor)
    draw.text((x, y - delta), text, font=font, fill=outlinecolor)
    draw.text((x, y + delta), text, font=font, fill=outlinecolor)

    # thicker border
    draw.text((x - delta, y - delta), text, font=font, fill=outlinecolor)
    draw.text((x + delta, y - delta), text, font=font, fill=outlinecolor)
    draw.text((x - delta, y + delta), text, font=font, fill=outlinecolor)
    draw.text((x + delta, y + delta), text, font=font, fill=outlinecolor)

    # now draw the text over it
    draw.text((x, y), text, font=font, fill=fillcolor)


def write_text(
    img, text, alignment="TL", fillcolor=(255, 255, 255), outlinecolor=(0, 0, 0), fontsize=0.03, padding=(5, 5)
):
    # (Image, str, str, Color, Color, Union[float, int], Union[float, Tuple[int, int]]) -> None

    if alignment not in {"TL", "TC", "TR", "BL", "BC", "BR"}:
        raise ValueError(f"Invalid alignment: {alignment}")

    if isinstance(fontsize, int):
        pass
    elif isinstance(fontsize, float):
        fontsize = int(img.height * fontsize)
    else:
        raise ValueError("fontsize must be float or int")

    if isinstance(padding, tuple):
        pass
    elif isinstance(padding, float):
        padding = int(img.width * padding), int(img.height * padding)
    else:
        raise ValueError("padding must be float or Tuple[int, int]")

    font = ImageFont.truetype("arial.ttf", fontsize)

    d = ImageDraw.Draw(img)
    size_text = d.textsize(text, font=font)

    if alignment == "TL":
        pos = padding
    elif alignment == "TC":
        pos = (img.width // 2 - size_text[0] // 2, padding[1])
    elif alignment == "TR":
        pos = (img.width - padding[0] - size_text[0], padding[1])
    elif alignment == "BL":
        pos = (padding[0], img.height - padding[1] - size_text[1])
    elif alignment == "BC":
        pos = (img.width // 2 - size_text[0] // 2, img.height - padding[1] - size_text[1])
    elif alignment == "BR":
        pos = (img.width - padding[0] - size_text[0], img.height - padding[1] - size_text[1])

    text_with_outline(d, pos, text, font, fillcolor, outlinecolor, 2)


def _fix_orientation(img, orientation):
    # type (Image, int) -> Image

    if orientation == 1:
        raise NoActionNeeded("File already properly rotated")
    elif orientation == 2:
        img = img.transpose(Image.FLIP_LEFT_RIGHT)
    elif orientation == 3:
        img = img.transpose(Image.ROTATE_180)
    elif orientation == 4:
        img = img.transpose(Image.FLIP_TOP_BOTTOM)
    elif orientation == 5:
        img = img.transpose(Image.TRANSPOSE)
    elif orientation == 6:
        img = img.transpose(Image.ROTATE_270)
    elif orientation == 7:
        img = img.transpose(Image.TRANSVERSE)
    elif orientation == 8:
        img = img.transpose(Image.ROTATE_90)
    else:
        raise ValueError("Unsupported orientation")

    return img


def fix_orientation(img, exif):
    # type (Image, dict) -> Image

    orientation = exif["0th"][piexif.ImageIFD.Orientation]
    img = _fix_orientation(img, orientation)
    exif["0th"][piexif.ImageIFD.Orientation] = 1

    return img



from __future__ import generator_stop

from typing import TYPE_CHECKING

import numpy as np

from .numpy import bincount_batch, unblock

if TYPE_CHECKING:
    from fractions import Fraction
    from typing import Tuple, Union

    Rational = Union[float, Fraction]


def grayscale(arr):
    return np.sum(arr, axis=-1) // arr.shape[-1]


def histogram_1d(arr, levels):
    # type: (np.ndarray, int) -> np.ndarray

    """Input shape of `arr`: [batch..., x]. Histogrammed over x and batched over the remaining dimensions.
    Output shape: [batch..., levels]
    """

    return bincount_batch(arr, -1, levels)


def resize_oar(max_width, max_height, dar):
    # type: (int, int, Rational) -> Tuple[int, int]

    maxdar = max_width / max_height

    if dar >= maxdar:  # wider than it should be
        width = max_width
        height = int(width / dar)
    else:  # thinner than it should be
        height = max_height
        width = int(height * dar)

    return width, height


def resize_maxsize(max_width, max_height, width, height):
    # type: (int, int, int, int) -> Tuple[int, int]

    return resize_oar(max_width, max_height, width / height)


def histogram_2d(arr, levels):
    # type: (np.ndarray, int) -> np.ndarray

    """Input shape of `arr`: [batch..., y, x]. Histogrammed over x and y and batched over
    the remaining dimensions.
    Output shape: [batch..., levels]
    """

    if len(arr.shape) < 2:
        raise ValueError("arr must be at least 2-dimensional")

    newshape = arr.shape[:-2] + (arr.shape[-2] * arr.shape[-1],)
    flattened = np.reshape(arr, newshape)

    return bincount_batch(flattened, -1, levels)


def block_histogram_2d(arr, by, bx, levels):
    # type: (np.ndarray, int, int, int) -> np.ndarray

    """Input shape of `arr`: [batch..., y, x]. Histogrammed over blocks of size bx and by
    and batched over the remaining dimensions.
    Output shape: [batch..., y/by, x/bx, levels]
    """

    invx = arr.shape[-1] // bx  # dimensions in unblock go from innerst to outerst
    invy = arr.shape[-2] // by
    blocks = unblock(arr, invx, invy)
    block_hists = histogram_1d(blocks, levels)
    return block_hists.reshape(arr.shape[:-2] + (invy, invx, -1))


def image_histogram(arr, levels=256):
    # type: (np.ndarray, int) -> np.ndarray

    """Input shape of `arr`: [batch..., x, y, channel]. It is summed over channels to create a grayscale
    image, then histogrammed over x and y and batched over the remaining dimensions.
    Output shape: [batch..., levels]
    """

    if len(arr.shape) < 3:
        raise ValueError("arr must be at least 3-dimensional")

    gray = grayscale(arr)
    return histogram_2d(gray, levels)


def image_block_histogram(arr, bx, by, levels=256):
    # type: (np.ndarray, int, int, int) -> np.ndarray

    """Input shape of `arr`: [batch..., x, y, channels]. It is summed over channels to create a grayscale
    image, then histogrammed over x and y and batched over the remaining dimensions.
    Output shape: [batch..., bx, by, levels]
    """

    if len(arr.shape) < 3:
        raise ValueError("arr must be at least 3-dimensional")

    gray = grayscale(arr)
    return block_histogram_2d(gray, bx, by, levels)



from __future__ import generator_stop

from typing import TYPE_CHECKING

import av

if TYPE_CHECKING:
    from typing import Iterator

    import numpy as np


class CorruptFile(Exception):
    pass


def iter_video(path):
    # type: (str, ) -> Iterator[np.ndarray]

    """Yield frames from a video file. For video device support see `genutility.cv.iter_video`."""

    with av.open(path, "r") as container:

        vs = container.streams.video[0]
        vs.thread_type = "AUTO"

        for frame in container.decode(vs):
            if frame.is_corrupt:
                raise CorruptFile("Frame %s at %s of %s is corrupt", frame.index, frame.time, path)
            yield frame.to_ndarray()  # or frame.to_rgb()



from __future__ import generator_stop

from datetime import datetime as _datetime
from datetime import timezone

try:
    _datetime.fromisoformat  # New in version 3.7
    datetime = _datetime

except AttributeError:

    import re
    from datetime import timedelta

    from ..string import tryint

    isoformatre = re.compile(r"^(\d{4})-(\d{2})-(\d{2})[ T](\d{2}):(\d{2}):(\d{2})(?:\.(\d{6}))?(\+|-)(\d{2}):(\d{2})$")
    isoformatre2 = re.compile(r"^(\d{4})-(\d{2})-(\d{2})$")

    class datetime(_datetime):
        @classmethod
        def fromisoformat(cls, date_string):
            # (str, ) -> datetime

            """Converts an ISO 8601 date and time with timezone string to a timezone aware datetime.
            eg. '2019-03-01T11:30:01.123000+00:00'
            eg. '2019-02-02T12:30:02.456000+04:00'
            eg. '2019-01-03T22:30:03.789000-07:00'
            """

            m = isoformatre.match(date_string)

            if m:
                groups = m.groups()
                year, month, day, hour, minute, second, milliseconds, tzsign, tzhours, tzminutes = map(tryint, groups)
            else:
                groups = isoformatre2.match(date_string).groups()
                year, month, day = map(tryint, groups)
                hour, minute, second, milliseconds, tzsign, tzhours, tzminutes = (0, 0, 0, 0, "+", 0, 0)

            offset = timedelta(hours=tzhours, minutes=tzminutes)

            if tzsign == "-":
                offset = -offset

            if milliseconds is None:
                milliseconds = 0

            return cls(year, month, day, hour, minute, second, milliseconds, tzinfo=timezone(offset))



from __future__ import generator_stop

import hashlib
import zlib
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

from .file import blockfileiter, read_file

if TYPE_CHECKING:
    from typing import Callable, Iterable, Optional, Union

    from _hashlib import HASH as Hashobj

    HashCls = Union[Callable[[], Hashobj], str]

FILE_IO_BUFFER_SIZE = 8 * 1024 * 1024


def hash_file(path, hashcls, chunksize=FILE_IO_BUFFER_SIZE, mode="rb", encoding=None):
    # type: (str, HashCls, int, str, Optional[str]) -> Hashobj

    if isinstance(hashcls, str):
        m = hashlib.new(hashcls)
    else:
        m = hashcls()
    for d in blockfileiter(path, mode, encoding, chunk_size=chunksize):
        m.update(d)
    return m


def hash_data(data, hashcls):
    # type: (bytes, HashCls) -> Hashobj

    """Hashes `data` with `hashcls`.
    `hashcls` can either be a string like "md5" or a hash object like `hashlib.md5`.
    """

    if isinstance(hashcls, str):
        m = hashlib.new(hashcls)
    else:
        m = hashcls()
    m.update(data)
    return m


md4_hash_data = partial(hash_data, hashcls="md4")
md5_hash_data = partial(hash_data, hashcls=hashlib.md5)
sha1_hash_data = partial(hash_data, hashcls=hashlib.sha1)


def crc32_hash_iter(it):
    # type: (Iterable[bytes], ) -> int

    """Create CRC32 hash from bytes takes from `it`."""

    prev = 0
    for data in it:
        prev = zlib.crc32(data, prev)

    return prev & 0xFFFFFFFF  # see https://docs.python.org/3/library/zlib.html#zlib.crc32


def crc32_hash_file(path, chunksize=FILE_IO_BUFFER_SIZE, mode="rb", encoding=None):
    # type: (str, int, str, Optional[str]) -> str

    """Return crc32 hash of file at `path`."""

    crcint = crc32_hash_iter(blockfileiter(path, mode, encoding, chunk_size=chunksize))
    return format(crcint, "08x")


md5_hash_file = partial(hash_file, hashcls=hashlib.md5)
sha1_hash_file = partial(hash_file, hashcls=hashlib.sha1)


def hashsum_file_format(hashobj, path):
    # type: (Hashobj, str) -> str

    return f"{hashobj.hexdigest()} *{path}"


ed2k_chunksize = 9728000


def ed2k_hash_file_v1(path):
    # type: (Path, ) -> str

    """Returns ed2k hash.
    This hashing method is used by
    - MLDonkey
    - Shareaza
    - HashCalc

    This differs from `ed2k_hash_file_v2` only if the file size is a multiple of `ed2k_chunksize`.
    """

    if path.stat().st_size <= ed2k_chunksize:
        return md4_hash_data(read_file(path, "rb")).hexdigest()

    ed2k_hashes = (md4_hash_data(data).digest() for data in blockfileiter(path, "rb", chunk_size=ed2k_chunksize))

    return md4_hash_data(b"".join(ed2k_hashes)).hexdigest()


def ed2k_hash_file_v2(path):
    # type: (Path, ) -> str

    """Returns ed2k hash.
    This hashing method is used by
    - eMule
    - AOM

    This differs from `ed2k_hash_file_v1` only if the file size is a multiple of `ed2k_chunksize`.
    """

    filesize = path.stat().st_size
    if filesize < ed2k_chunksize:
        return md4_hash_data(read_file(path, "rb")).hexdigest()

    ed2k_hashes = list(md4_hash_data(data).digest() for data in blockfileiter(path, "rb", chunk_size=ed2k_chunksize))

    if filesize % ed2k_chunksize == 0:
        ed2k_hashes.append(md4_hash_data(b"").digest())

    return md4_hash_data(b"".join(ed2k_hashes)).hexdigest()



from __future__ import generator_stop

from collections.abc import Iterable, Sized
from typing import TYPE_CHECKING, Any, Optional, Sequence, TypeVar, Union

if TYPE_CHECKING:

    from typing_extensions import Protocol

    Number = Union[int, float]
    T = TypeVar("T")

    class SizedIterable(Iterable[T], Sized):
        pass

    class Comparable(Protocol):
        def __eq__(self, other):
            # type: (Any, ) -> bool
            ...

        def __ne__(self, other):
            # type: (Any, ) -> bool
            ...

    class Orderable(Protocol):
        def __lt__(self, other):
            # type: (Any, ) -> bool
            ...

        def __gt__(self, other):
            # type: (Any, ) -> bool
            ...

        def __le__(self, other):
            # type: (Any, ) -> bool
            ...

        def __ge__(self, other):
            # type: (Any, ) -> bool
            ...

    class Computable(Protocol):
        def __neg__(self):
            # type: () -> Any
            ...

        def __add__(self, other):
            # type: (Any, ) -> Any
            ...

        def __sub__(self, other):
            # type: (Any, ) -> Any
            ...

        def __mul__(self, other):
            # type: (Any, ) -> Any
            ...

        def __truediv__(self, other):
            # type: (Any, ) -> Any
            ...

        def __floordiv__(self, other):
            # type: (Any, ) -> Any
            ...

        def __abs__(self):
            # type: () -> Any
            ...

    class MutableComputable(Protocol):
        def __neg__(self):
            # type: () -> Any
            ...

        def __add__(self, other):
            # type: (Any, ) -> Any
            ...

        def __iadd__(self, other):
            # type: (Any, ) -> Any
            ...

        def __sub__(self, other):
            # type: (Any, ) -> Any
            ...

        def __isub__(self, other):
            # type: (Any, ) -> Any
            ...

        def __mul__(self, other):
            # type: (Any, ) -> Any
            ...

        def __imul__(self, other):
            # type: (Any, ) -> Any
            ...

        def __truediv__(self, other):
            # type: (Any, ) -> Any
            ...

        def __itruediv__(self, other):
            # type: (Any, ) -> Any
            ...

        def __floordiv__(self, other):
            # type: (Any, ) -> Any
            ...

        def __ifloordiv__(self, other):
            # type: (Any, ) -> Any
            ...

    class Cursor(Protocol):
        def callproc(self, procname, parameters):
            # type: (str, Optional[Sequence[Any]]) -> None
            ...

        def close(self):
            # type: () -> None
            ...

        def execute(self, operation, parameters):
            # types: (str, Optional[Sequence[Any]]) -> None
            ...

        def executemany(self, operation, seq_of_parameters):
            # types: (str, Sequence[Sequence[Any]]) -> None
            ...

        def fetchone(self):
            # types: () -> Optional[Sequence[Any]]
            ...

        def fetchmany(self, size):
            # types: (Optional[int], ) -> Sequence[Sequence[Any]]
            ...

        def fetchall(self):
            # types: () -> Sequence[Sequence[Any]]
            ...

        def nextset(self):
            # type: () -> Optional[bool]
            ...

        def setinputsizes(self, sizes):
            # types: (Sequence[int], ) -> None
            ...

        def setoutputsize(self, size, column):
            # type: (int, Optional[int]) -> None
            ...

    class Connection(Protocol):
        def close(self):
            # type: () -> None
            ...

        def commit(self):
            # type: () -> None
            ...

        def rollback(self):
            # type: () -> None
            ...

        def cursor(self):
            # type: () -> Cursor
            ...



from __future__ import generator_stop

from typing import TYPE_CHECKING

from .math import argmax, argmin

if TYPE_CHECKING:
    from typing import MutableSequence


def bubble_sort(seq):
    # type: (MutableSequence, ) -> None

    """Slightly optimized version of stable bubble sort. Sorts input `seq` in place."""

    n = len(seq)
    while True:
        newn = 1
        for i in range(1, n):
            if seq[i - 1] > seq[i]:
                seq[i - 1], seq[i] = seq[i], seq[i - 1]
                newn = i
        if newn == 1:
            break
        n = newn


def selection_sort_min(seq):
    # inplace, unstable

    n = len(seq)
    for i in range(n - 1):
        m = argmin(seq, i, n)
        if m != i:
            seq[i], seq[m] = seq[m], seq[i]


def selection_sort_max(seq):
    # inplace, unstable

    for n in range(len(seq) - 1, 0, -1):
        m = argmax(seq[: n + 1])
        if m != n:
            seq[m], seq[n] = seq[n], seq[m]


def selection_sort_ll(seq):
    # inplace, stable, on linked lists
    raise NotImplementedError



from __future__ import generator_stop

STARS = {
    "Sun": {
        "mass": 1.9891e30,
    }
}

PLANETS = {
    "Earth": {
        "mass": 5.9736e24,
    }
}

UNITS = {
    "Astronomical_unit": 149597870700,  # m
    "Day": 60 * 60 * 24,  # s
}

FUNDAMENTAL = {
    "Gravitational_constant": 6.67384e-11,
}



from __future__ import generator_stop

from itertools import chain

from setuptools import setup

# nltk.download("punkt")

with open("README.md", encoding="utf-8") as fr:
    long_description = fr.read()

extras_require = {
    "args": [],
    "algorithms": [],
    "atomic": [],
    "binary": [],
    "concurrency": [],
    "constants_physics": [],
    "constants_video": [],
    "csv": [],
    "datetime": [],
    "debug": [],
    "dict": [],
    "datastructures": [],
    "error": [],
    "exceptions": [],
    "factorial": [],
    "func": [],
    "file": [],
    "filesystem": [],
    "geometry": [],
    "hash": [],
    "iter": [],
    "imdb": [],
    "indexing": [],
    "latex": [],
    "math": [],
    "metrictree": [],
    "mongo": [],
    "object": [],
    "ops": [],
    "os": ["ctypes-windows-sdk>=0.0.7; sys_platform == 'win32'"],
    "pickle": [],
    "profile": [],
    "rand": [],
    "ringlist": [],
    "search": [],
    "sequence": [],
    "set": [],
    "signal": [],
    "sort": [],
    "sql": [],
    "sqlite": [],
    "statistics": [],
    "stdio": [],
    "string": [],
    "test": [],
    "time": [],
    "typing": ["typing-extensions"],
    "text": [],
    "twitch": [],
    "url": [],
    # "browscap": ["re2"], # unfinished
    "cholesky": ["numpy", "numba"],
    "config": ["jsonschema", "sortedcontainers"],
    "downloadmanager": ["aiohttp>=0.16.3", "orderedset"],
    "encoder": ["numpy", "nltk"],
    "filesdb": ["tls-property"],
    "fingerprinting": ["numpy", "numba", "pillow"],
    "html": ["beautifulsoup4"],
    "http": ["requests", "certifi"],
    "http2": ["hyper"],
    "image": ["numpy"],
    "json": ["jsonschema"],
    "lda": ["numpy", "nltk"],
    "metrics": ["numpy"],
    "mp4": ["setuptools"],
    "multihash": ["rhash"],
    "net": ["netifaces"],
    "nlp": ["nltk"],
    "regression": ["numpy"],
    "resizing": ["numpy", "opencv-python", "scikit-learn"],
    "rich": ["rich"],
    "scene_change_detection": ["numpy"],
    "sudoku": ["setuptools"],
    "tls": ["cryptography", "pyOpenSSL>=17.5.0"],
    "torrent": ["bencode.py>=2.0.0"],  # bencoding ?
    "videofile": ["opencv-python", "av"],
    "win32": ["wmi", "pypiwin32"],
    "widgets": ["wxPython"],
    "xsl": ["lxml"],
    # helpers for specific packages
    "aria": ["aria2p", "requests"],
    "av": ["av"],
    "cv": ["opencv-python", "numpy"],
    "flask": ["flask", "werkzeug>=0.11.11"],
    "gensim": ["gensim>=3.3.0", "numpy"],
    "msgpack": ["msgpack>=0.6.0"],
    "mediainfo": ["pymediainfo"],
    "numpy": ["numpy"],
    "numba": ["numba"],
    "nltk": ["nltk"],
    "networkx": ["networkx"],
    "pandas": ["pandas"],
    "pdf": ["PyPDF2"],
    "pillow": ["pillow", "piexif"],
    "powerpoint": ["python-pptx"],
    "rasa": ["requests", "aiohttp>=2.3", "ruamel.yaml"],
    "salesforce": ["simple-salesforce>=1.1.0", "simplejson", "aiohttp>=2.3", "requests", "pandas"],
    "scipy": ["numpy", "scipy"],
    "spark": ["pyspark>=3.0.0"],
    "toml": ["toml"],
    "torch": ["torch"],
    "twisted": ["twisted"],
    "tensorflow": ["tensorflow"],
    "yaml": ["ruamel.yaml"],
}

extras_require["all"] = sorted(set(chain.from_iterable(extras_require.values())))

setup(
    author="Dobatymo",
    name="genutility",
    version="0.0.79",
    url="https://github.com/Dobatymo/genutility",
    description="A collection of various Python utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
    ],
    packages=["genutility", "genutility.compat", "genutility.fileformats", "genutility.hardware", "genutility.win"],
    package_data={"genutility": ["py.typed", "data/*.tsv"]},
    python_requires=">=3.6",
    install_requires=["ctypes-windows-sdk>=0.0.7; sys_platform=='win32'", "typing-extensions"],
    extras_require=extras_require,
    license_files=["LICENSE"],
)



from __future__ import generator_stop

from math import exp, sqrt
from typing import Callable, Dict, Generic, Iterator, Optional, Tuple, TypeVar

import numpy as np

# from .numba import opjit

T = TypeVar("T")

RGB_YELLOW = (255, 255, 0)
RGB_WHITE = (255, 255, 255)


def assert_square(name, value):
    # type: (str, np.ndarray) -> None

    """Raises a ValueError if matrix `value` is not square.
    value: [A, A]
    """

    if not len(value.shape) == 2 or value.shape[0] != value.shape[1]:
        raise ValueError(f"{name} must be a square")


def normalized_choice(
    p_ind, p_prob
):  # is this neccessary? I think `np.random.choice` can handle unnormalized probabilities
    return np.random.choice(p_ind, p=p_prob / np.sum(p_prob))


def shannon_entropy(ps, base=2):
    # type: (np.ndarray, int) -> float

    """Calculates the Shannon entropy for probabilities `ps` with `base`."""

    return -np.sum(ps * np.log(ps) / np.log(base))


def is_rgb(img):
    # type: (np.ndarray, ) -> bool

    return len(img.shape) >= 1 and img.shape[-1] == 3


# @opjit() # np.errstate doesn't work in numba...
def rgb_to_hsi(image):
    # type: (np.ndarray, ) -> np.ndarray

    """Converts an array [..., channels] of RGB values to HSI color values (H in rad).
    RGB values are assumed to be normalized to (0, 1).
    """

    if not is_rgb(image):
        raise ValueError("Input needs to be an array of RGB values")

    r = image[..., 0]
    g = image[..., 1]
    b = image[..., 2]

    out = np.zeros_like(image)

    # allequal = (img == img[:, :, 0, np.newaxis]).all(axis=-1)

    with np.errstate(invalid="ignore"):
        tmp = (2.0 * r - g - b) / 2.0 / np.sqrt((r - g) ** 2 + (r - b) * (g - b))  # if r==g==b then 0/0

        theta = np.arccos(np.clip(tmp, -1.0, +1.0))

        out[..., 0] = np.where(b <= g, theta, 2 * np.pi - theta)  # H
        out[..., 2] = np.sum(image, axis=-1) / 3.0  # I
        out[..., 1] = 1 - np.amin(image, axis=-1) / out[..., 2]  # S if r==g==b==0 then 0/0

    np.nan_to_num(out[..., 0:2], copy=False)

    return out


# @opjit() # np.dot with more than 2 dimensions is not supported...
def rgb_to_ycbcr(image):
    # type: (np.ndarray, ) -> np.ndarray

    """Converts an array [..., channels] of RGB values to Digital Y'CbCr (0-255).
    RGB values are assumed to be normalized to (0, 1).
    Don't forget to cast to uint8 for pillow.
    """

    """  from RGB (0-1).
    """

    if not is_rgb(image):
        raise ValueError("Input needs to be an array of RGB values")

    m = np.array(
        [
            [+065.481, +128.553, +024.966],
            [-037.797, -074.203, +112.000],
            [+112.000, -093.786, -018.214],
        ]
    )
    a = np.array([16, 128, 128])

    return np.dot(image, m.T) + a


def random_triangular_matrix(size, lower=True):
    # type: (int, bool) -> np.ndarray

    """Returns a triangular matrix with random value between 0 and 1 uniformly."""

    a = np.random.uniform(0, 1, (size, size))
    if lower:
        ind = np.triu_indices(5, 1)
    else:
        ind = np.tril_indices(5, 1)
    a[ind] = 0

    return a


def is_square(A):
    # type: (np.ndarray, ) -> bool

    if len(A.shape) != 2:
        return False

    if A.shape[0] != A.shape[1]:
        return False

    return True


def batch_vTAv(A, v):
    # type: (np.ndarray, np.ndarray) -> np.ndarray

    """Performs batched calculation of `v^T A v` transform.
    Special case of bilinear form `x^T A y`

    A: float[B, X, X]
    v: float[B, X]

    returns: float[B]
    """

    """ Faster than
        Av = np.matmul(A, v[...,:,None]) # [B, X, 1]
        return np.matmul(v[...,None,:], Av).squeeze((-2, -1)) # [B]
    """

    return np.einsum("...k,...kl,...l->...", v, A, v)


def batch_inner(a, b, verify=True):
    # type: (np.ndarray, np.ndarray, bool) -> np.ndarray

    """Performs a batched inner product over the last dimension.
    Replacement for deprecated `from numpy.core.umath_tests import inner1d`.
    Shapes: (B, X), (B, X) -> (B, )
    """

    if verify and a.shape != b.shape:
        raise ValueError("All dimensions have to be equal")

    if a.shape[-1] == 0:
        return np.empty_like(a)

    return np.einsum("...i,...i->...", a, b)  # faster than np.sum(a * b, axis=-1)


def batch_outer(a, b, verify=True):
    # type: (np.ndarray, np.ndarry, bool) -> np.ndarray

    """Performs a batched outer product over the last dimension.
    Shapes: (B, X), (B, Y) -> (B, X, Y)
    """

    if verify and a.shape[:-1] != b.shape[:-1]:
        raise ValueError("All except the last dimension have to be equal")

    return np.einsum("...i,...j->...ij", a, b)  # slightly faster than np.multiply(a[...,:,None], b[...,None,:])


def batchtopk(probs, k=None, axis=-1, reverse=False):
    # type: (np.ndarray, np.ndarray, int, bool) -> np.ndarray

    """`probs` values ndarray
    `k` take the smallest `k` elements, if `reverse` is False
            and the largest `k` if `reverse` is True
    `axis` sorting and selection axis.
    """

    if k is not None and k <= 0:
        raise ValueError("k must be larger than 0. Use None to chose all elements.")

    if axis != -1:
        raise ValueError("Only last axis supported atm")

    if len(probs.shape) <= 1:
        raise ValueError("probs must be at least 2-dimensional")

    if reverse:
        sign = -1
    else:
        sign = 1

    indices = np.argsort(sign * probs, axis=-1)  # use argpartition?
    probs = np.take_along_axis(probs, indices[..., :k], axis=-1)

    return indices, probs


# @opjit() # np.diagonal not supported by numba as of version 0.52.0
def logtrace(m):
    """Calcuates the sum of the logs of the diagonal elements (batchwise if necessary)
    m: [..., x, x]
    """

    """ note: performance cannot easily be improve by numba.
        `np.diagonal` not supported by numba 0.52.0
    """

    return np.sum(np.log(np.diagonal(m, axis1=-2, axis2=-1)), axis=-1)


def shiftedexp(pvals):
    # type: (np.ndarray, ) -> np.ndarray

    """Shifts `pvals` by the largest value in the last dimension
    before the exp is calculated to prevent overflow (batchwise if necessary).
    Can be used if probabilities are normalized again later.
    """

    if pvals.shape[-1] == 0:
        return np.empty_like(pvals)

    return np.exp(pvals - np.amax(pvals, axis=-1)[..., None])


class Sampler:

    """Sample from discrete CDF."""

    def __init__(self, cdf):
        self.cdf = cdf
        self.psum = cdf[-1]

    def __call__(self):
        # type: () -> int

        """Sample one."""

        rand = np.random.uniform(0, self.psum)
        return np.searchsorted(self.cdf, rand, side="right")

    def sample(self, n):
        # type: (int, ) -> np.ndarray

        """Sample `n`."""

        rands = np.random.uniform(0, self.psum, n)
        return np.searchsorted(self.cdf, rands, side="right")

    def pdf(self, n, minlength=None):
        # type: (int, Optional[int]) -> np.ndarray

        out = np.bincount(self.sample(n), minlength=minlength)
        return out / n


def sample_probabilities(pvals):
    # type: (np.ndarray, ) -> Callable[[], int]

    """Sample from list of probabilities `pvals` with replacement.
    The probabilities don't need to be normalized.
    """

    return Sampler(np.cumsum(pvals))


class UnboundedSparseMatrix(Generic[T]):
    def __init__(self, dtype=float):
        # type: (type, ) -> None

        self.dtype = dtype
        self.zero = self.dtype(0)
        self.m = dict()  # type: Dict[Tuple[int, int], T]
        self.cols = 0
        self.rows = 0

    def __getitem__(self, slice):
        # type: (Tuple[int, int], ) -> T

        return self.m.get(slice, self.zero)

    def __setitem__(self, slice, value):
        # type: (Tuple[int, int], T) -> None

        c, r = slice
        self.cols = max(self.cols, c + 1)
        self.rows = max(self.rows, r + 1)
        self.m[slice] = value

    def todense(self):
        # type: () -> np.ndarray

        ret = np.zeros((self.cols, self.rows), dtype=self.dtype)

        for slice, value in self.m.items():
            ret[slice] = value

        return ret


def normalize(pvals):
    # type: (np.ndarray, ) -> np.ndarray

    return pvals / np.sum(pvals)


def categorical(pvals):
    # type: (np.ndarray, ) -> int

    """Sample from the categorical distribution using `pvals`.
    See: https://en.wikipedia.org/wiki/Categorical_distribution
    """

    return sample_probabilities(pvals)()  # faster than: np.argmax(np.random.multinomial(1, normalize(pvals)))


def population2cdf(population):
    # type: (np.ndarray, ) -> np.ndarray

    """Convert a population (list of observations) to a CDF."""

    population = np.sort(population)
    return np.searchsorted(population, population, side="right") / len(population)


def pmf2cdf(pdf):
    # type: (np.ndarray, ) -> np.ndarray

    """Convert a discrete PDF into a discrete CDF."""

    cdf = np.cumsum(pdf)
    return cdf / cdf[-1]


def _two_sample_kolmogorov_smirnov_same_length(cdf1, cdf2, n1, n2):
    # type: (np.ndarray, np.ndarray, int, int) -> Tuple[float, float]

    # note: yields different results as `scipy.stats.ks_2samp`

    if len(cdf1) != len(cdf2):
        raise ValueError("Both CDFs must have the same length")

    D = np.amax(np.abs(cdf1 - cdf2))  # K-S statistic
    level = exp(-2 * (D / sqrt((n1 + n2) / (n1 * n2))) ** 2)
    return D, level


def _two_sample_kolmogorov_smirnov_population(p1, p2, alpha=0.05):
    # type: (np.ndarray, np.ndarray, float) -> Tuple[float, float, bool]

    # note: yields different results as `scipy.stats.ks_2samp`

    cdf1 = population2cdf(p1)
    cdf2 = population2cdf(p2)

    statistic, pvalue = _two_sample_kolmogorov_smirnov_same_length(cdf1, cdf2, len(cdf1), len(cdf2))
    reject = pvalue < alpha
    return statistic, pvalue, reject


def _two_sample_kolmogorov_smirnov_pmf(pmf1, pmf2, alpha=0.05):
    # note: yields different results as `scipy.stats.ks_2samp`
    """Tests the null hypothesis that both samples belong to the same distribution.
    See: https://en.wikipedia.org/wiki/Kolmogorov%E2%80%93Smirnov_test#Two-sample_Kolmogorov%E2%80%93Smirnov_test
    """

    cdf1 = np.cumsum(pmf1)
    cdf2 = np.cumsum(pmf2)

    n1 = cdf1[-1]
    n2 = cdf2[-1]

    # cannot be inplace because of type conversion
    cdf1 = cdf1 / n1
    cdf2 = cdf2 / n2

    statistic, pvalue = _two_sample_kolmogorov_smirnov_same_length(cdf1, cdf2, n1, n2)
    reject = pvalue < alpha
    return statistic, pvalue, reject


def inf_matrix_power(pm):
    # type: (np.ndarray, ) -> np.ndarray

    """Calculate stochastic matrix `pm` to the power of infinity,
    by finding the eigenvector which corresponds to the eigenvalue 1.
    """

    w, v = np.linalg.eig(
        pm
    )  # scipy.linalg.eig would probably by faster as it can return the left and right eigen vectors

    if not np.isclose(w[0], 1.0):
        raise ValueError("The first eigenvalue is not none. Is this a right stochastic matrix?")

    vi = np.linalg.inv(v)
    d = np.zeros(pm.shape[0], dtype=np.float)
    d[0] = 1.0

    return np.matmul(v, np.matmul(np.diag(d), vi))


def decompress(selectors, data, default):
    # type: (np.ndarray[bool], np.ndarray[T], T) -> None

    """Same result as:
    `np.array(list(genutility.iter.decompress(selectors, iter(data), default)))`
    but faster
    """

    out = np.full(len(selectors), default)
    out[selectors] = data
    return out


def unblock(arr, n1, n2, axis1=-1, axis2=-2, blocksize=False):

    """Inverse of np.block.
    Set axis to (-2, -1) to modify the order of the result.
    """

    """ test (stackoverflow): Ok, so considering I have N block matrices with bm x bn dimension and want to stack them in a m x n matrix, provided N = m x n, I would then have x.reshape(m,n,bm,bn).swapaxes(1,2).reshape(bm*m,-1)
    """

    s = np.array(arr.shape)
    if s[axis1] % n1 != 0 or s[axis2] % n2 != 0:
        raise ValueError(f"{s[axis1]}x{s[axis2]} does not divide by {n1}x{n2}")

    if blocksize:
        n1 = s[axis1] // n1
        n2 = s[axis2] // n2

    arr = np.array(np.split(arr, n1, axis1))
    arr = np.array(np.split(arr, n2, axis2))

    inv_blocksize = n1 * n2
    total = s[axis1] * s[axis2]
    s[axis2] = inv_blocksize
    s[axis1] = total // inv_blocksize

    return np.reshape(arr, s)


def unblock2d(arr: np.ndarray, n1: int, n2: int, blocksize: bool = False) -> np.ndarray:
    """
    arr: (width, height)

    returns: (block_num, block_size)
    """

    s = np.array(arr.shape)
    if blocksize:
        bs1 = n1
        bs2 = n2
        bn1 = s[0] // n1
        bn2 = s[1] // n2
    else:
        bn1 = n1
        bn2 = n2
        bs1 = s[0] // n1
        bs2 = s[1] // n2

    return arr.reshape(bn1, bs1, bn2, bs2).swapaxes(1, 2).reshape(bn1 * bn2, bs1 * bs2)


def block2d(arr: np.ndarray, bn1: int, bn2: int, bs1: int, bs2: int) -> np.ndarray:
    """
    arr: (block_num, block_size)

    returns: (width, height)
    """

    s = np.array(arr.shape)
    assert s[0] == bn1 * bn2
    assert s[1] == bs1 * bs2

    return arr.reshape(bn1, bn2, bs1, bs2).swapaxes(1, 2).reshape(bn1 * bs1, bn2 * bs2)


def remove_color(img, ratio, neutral_color=RGB_WHITE):
    # type: (np.ndarray, float, Tuple[int, int, int]) -> np.ndarray

    """Replace colored pixels with a `neutral_color`. The `ratio` defines the 'colorfulness' above
    which level the pixel should be replace.
    I.e. if the `ratio` is 1 nothing will be replaced,
    if `ratio` is 0 only strict greys are kept unmodified.
    """

    channels = img.shape[-1]
    assert channels == 3, "Not a 3 channel color image"

    norm = np.std(np.array(RGB_YELLOW))  # this is the same for all pure colors

    sd = np.std(img, axis=-1)
    img[sd > ratio * norm] = neutral_color


# @opjit(cache=False) doesn't help a lot, because it's a generator
def sliding_window_2d(image, window_size, step_size=(1, 1)):
    # type: (np.ndarray, Tuple[int, int], Tuple[int, int]) -> Iterator[np.ndarray]

    win_x, win_y = window_size
    step_x, step_y = step_size
    height, width = image.shape[0:2]

    for y in range(0, height - win_y + 1, step_y):
        for x in range(0, width - win_x + 1, step_x):
            yield image[y : y + win_y, x : x + win_x, ...]


def histogram_correlation(hist1, hist2):
    # type: (np.ndarray, np.ndarray) -> np.ndarray

    """Input shape of `hist1` and `hist2`: [batch..., levels]. The correlation is calculated over `levels`
    and then batched over the remaining dimensions.
    """

    assert len(hist1.shape) >= 1 and hist1.shape == hist2.shape

    h1norm = (hist1.T - np.mean(hist1, axis=-1)).T
    h2norm = (hist2.T - np.mean(hist2, axis=-1)).T

    num = np.sum(h1norm * h2norm, axis=-1)
    denom = np.sqrt(np.sum(h1norm ** 2, axis=-1) * np.sum(h2norm ** 2, axis=-1))
    return num / denom


def bincount_batch(x, axis=-1, minlength=0):
    # type: (np.ndarray, int, int) -> np.ndarray

    if x.shape[axis] == 0:
        raise ValueError("Specified axis of x cannot be 0")

    minlength = max(minlength, x.max() + 1)
    return np.apply_along_axis(np.bincount, axis, x, minlength=minlength)



from __future__ import generator_stop

import re
import socket
from typing import TYPE_CHECKING

import netifaces

if TYPE_CHECKING:
    from typing import Optional


def is_ipv4(s):
    # type: (str, ) -> bool

    """Tests if `s` is a IPv4 address string.
    It only validates the common `x.x.x.x` format and rejects less common ones like `x.x`.
    """

    try:
        nums = tuple(map(int, s.split(".")))
        return len(nums) == 4 and min(nums) >= 0 and max(nums) <= 255
    except (AttributeError, ValueError):
        return False


simple_email_regex = r"[^@\s]+@[^@\s]+\.[^@\s]+"


def is_email(s):
    # type: (str, ) -> bool

    """Coarse check to test of a string is an email. Will accept some invalid ones like
    "asd@.asd.com" and reject some valid ones like "asd@localhost"
    """

    return re.fullmatch(simple_email_regex, s) is not None


def get_standard_gateway(default=None):
    # type: (Optional[str], ) -> Optional[str]

    """Returns the standard IPv4 gateway."""

    for ipv4_gateway in netifaces.gateways()["default"][netifaces.AF_INET]:
        try:
            socket.inet_aton(ipv4_gateway)
            return ipv4_gateway
        except OSError:
            pass

    return default



from __future__ import generator_stop

from typing import TYPE_CHECKING

from werkzeug.exceptions import BadRequest
from werkzeug.wsgi import get_host

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Iterable, Mapping, Optional

    WsgiApp = Any


class HostDispatcher:

    """Dispatch WSGI apps by host.
    app = HostDispatcher({'a.com': app_a, 'b.com': app_b})
    """

    def __init__(self, hosts, default=None):
        # type: (Dict[str, WsgiApp], Optional[WsgiApp]) -> None

        self.hosts = hosts
        self.default = default or BadRequest()  # rfc7230#section-5.4

    def __call__(self, environ, start_response):
        # type: (Mapping, Callable) -> Iterable[bytes]

        host = get_host(environ)
        app = self.hosts.get(host, self.default)
        return app(environ, start_response)



from __future__ import generator_stop

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from typing import Callable, Iterable, List, Optional

    Tokenizer = Callable[[str], Iterable[str]]


def hamming_distance(a, b):
    # type: (bytes, bytes) -> int

    a = np.unpackbits(np.frombuffer(a, dtype=np.uint8))
    b = np.unpackbits(np.frombuffer(b, dtype=np.uint8))
    return np.count_nonzero(a != b)


def default_tokenizer(text):
    # type: (str, ) -> List[str]

    return text.lower().split()


def same_words_similarity(a, b, tokenizer=None):
    # type: (str, str, Optional[Tokenizer]) -> int

    tokenizer = tokenizer or default_tokenizer

    set_a = set(tokenizer(a))
    set_b = set(tokenizer(b))

    return len(set_a & set_b)



from __future__ import generator_stop

import os.path
from configparser import ConfigParser
from importlib.util import find_spec
from typing import TYPE_CHECKING

from .os import get_appdata_dir

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Union


def sort_config(inpath, outpath):
    # type: (str, str) -> None

    from sortedcontainers import SortedDict

    config = ConfigParser(dict_type=SortedDict)
    config.read(inpath, encoding="utf-8")
    with open(outpath, "wt", encoding="utf-8") as fw:
        config.write(fw)


def _load(name):
    # type: (str, ) -> Dict[str, Any]

    from .toml import read_toml

    configfilename = name + ".toml"

    # try appdata directory
    try:
        return read_toml(os.path.join(get_appdata_dir(), name, configfilename))
    except FileNotFoundError:
        pass

    # try module directory
    try:
        spec = find_spec(name)

        if spec is None:
            raise ImportError(f"No module named '{name}'")  # or FileNotFoundError?

        if spec.has_location:
            assert spec.origin  # for mypy
            modpath = os.path.dirname(spec.origin)
        else:
            try:
                modpath = spec.submodule_search_locations[0]  # type: ignore
            except (TypeError, IndexError):
                raise FileNotFoundError

        return read_toml(os.path.join(modpath, configfilename))
    except (ImportError, FileNotFoundError):
        pass

    # try working directory
    try:
        return read_toml(configfilename)
    except FileNotFoundError:
        raise FileNotFoundError(
            f"{configfilename} could not be found in application data, module path or current directory"
        )


def load(name, json_schema=None):
    # type: (str, Optional[Union[dict, str]]) -> Dict[str, Any]

    from .json import read_json_schema

    obj = _load(name)

    if json_schema is None:
        return obj

    from jsonschema import validate

    if isinstance(json_schema, str):
        json_schema = read_json_schema(json_schema)

    validate(obj, json_schema)
    return obj



from __future__ import generator_stop

from typing import TYPE_CHECKING

try:
    from contextlib import nullcontext  # pylint: disable=no-name-in-module # New in version 3.7

except ImportError:

    if TYPE_CHECKING:
        from typing import Optional, TypeVar

        T = TypeVar("T")

    try:
        from contextlib import AbstractContextManager
    except ImportError:
        AbstractContextManager = object

    class nullcontext(AbstractContextManager):

        """A context manager which doesn't do anything except return its argument.
        Kind of like an identity function for contexts, useful as default argument.
        """

        def __init__(self, enter_result=None):
            # type: (Optional[T], ) -> None

            self.enter_result = enter_result

        def __enter__(self):
            # type: () -> Optional[T]

            return self.enter_result

        def __exit__(self, exc_type, exc_value, traceback):
            pass
